package flowsim.core;

import com.sun.java.util.collections.Collection;
import com.sun.java.util.collections.Iterator;
import com.sun.java.util.collections.LinkedList;

final public class DataTransferImpl implements DataTransfer, FlowMonitor
{
	final static boolean DEBUG=false;

	final Simulator s;
	final int from_port;
	final int to_port;
	final UnicastFlow flow;
	final int total_bits;
	final Collection monitors;
	long last_time;
	int remaining_bits;
	EventQueueElement transmitted_event;
	int bw, oldbw;
	long start_time;

	final class TransmittedEvent implements Event
	{
		public void trigger() { transmitted(); }
	}

	final class ArrivedEvent implements Event
	{
		public void trigger() { arrived(); }
	}

	public long get_start_time() {
		return start_time;
	}

	int get_total_delay() {
		int delay=0;
		for(UnicastFlow uf = flow; uf!=null; uf = uf.get_child()) {
			LinkManager lm = uf.get_iface();
			if(lm!=null) delay+=lm.get_link().get_delay();
		}
		return delay;
	}

	void transmitted() {
		flow.stop();
		s.schedule(get_total_delay(),new ArrivedEvent());

		// monitors
		Iterator i = monitors.iterator();
		while(i.hasNext()) {
			DataTransferMonitor m = (DataTransferMonitor) i.next();
			m.datatransfer_transmitted(this);
		}
	}

	void arrived() {
		// monitors
		Iterator i = monitors.iterator();
		while(i.hasNext()) {
			DataTransferMonitor m = (DataTransferMonitor) i.next();
			m.datatransfer_arrived(this);
		}
	}
	
	public final int get_from_port() {
		return from_port;
	}

	public final int get_to_port() {
		return to_port;
	}

	public final int get_total_bits() {
		return total_bits;
	}

	public final Flow get_flow() {
		return flow;
	}

	public final int get_bw() {
		return bw;
	}

	public final void attach_monitor(DataTransferMonitor monitor) {
		monitors.add(monitor);
	}

	public final void detach_monitor(DataTransferMonitor monitor) {
		monitors.remove(monitor);
	}

	public void start() {
		if(DEBUG) s.debug(this,"Started");
		flow.start();
		create_transmitted_event();
		last_time = start_time = s.now();

		// monitors 
		Iterator i = monitors.iterator();
		while(i.hasNext()) {
			DataTransferMonitor m = (DataTransferMonitor) i.next();
			m.datatransfer_started(this);
		}
	}

	public void abort() {
		transmitted_event.remove();
		flow.stop();
	}

	public final int get_remaining_bits() {
		return remaining_bits;
	}

	void create_transmitted_event() {
		long t;

		if(bw>0) {
			t = (long) 1000*1000*remaining_bits/bw;
		}
		else {
			t = Scheduler.TIME_INFINITY;
		}
		
		if(DEBUG) s.debug(this,"creating TransmittedEvent (rbits="+
				remaining_bits+" ,bw="+bw+" -> t="+t+").");

		Event event;
		if(transmitted_event!=null) {
			event=transmitted_event.event;
			transmitted_event.remove();
		}
		else {
			event = new TransmittedEvent();
		}
		transmitted_event = s.schedule(t,event);
	}

	public void flow_changed_bw(Flow f, int _oldbw) {
		oldbw = bw;
		bw = f.get_bw();

		if(DEBUG) s.debug(this,"flow_changed_bw -> bw="+bw);

		if(transmitted_event!=null) {
			long now = s.now();
			if(DEBUG) s.debug(this,"set_bw ("+
					"remaining_bits="+remaining_bits+", "+
					"last_time="+last_time+", "+
					"now="+now+", "+
					"last_bw="+oldbw+", "+
					"bw="+flow.get_bw());
			remaining_bits -= (now-last_time)*oldbw/1000000;
			if(remaining_bits<0) remaining_bits=0;
			create_transmitted_event();
			last_time = s.now();
		}

		// monitors 
		Iterator i = monitors.iterator();
		while(i.hasNext()) {
			DataTransferMonitor m = (DataTransferMonitor) i.next();
			m.datatransfer_changed_bw(this);
		}
	}

	public void flow_started(Flow f) {}
	public void flow_stopped(Flow f) {}

	public void flow_blocked(Flow f) {
		abort();
		
		// monitors 
		Iterator i = monitors.iterator();
		while(i.hasNext()) {
			DataTransferMonitor m = (DataTransferMonitor) i.next();
			m.datatransfer_blocked(this);
		}
	}

	public void flow_changed_desired_bw(Flow f, int was_bw) {}
	public void flow_changed_available_bw(Flow f, int was_bw) {}

	public final String toString() {
		return "DataTransfer("+Integer.toHexString(hashCode())+":"+
			flow.get_from()+","+flow.get_to()+")";
	}

	/**
	 * @param total_bits Size of the data-transfer in bits.
	 */
	public DataTransferImpl(Simulator s, int from_port, int to_port,
		int total_bits, UnicastFlow flow)
	{
		this.s = s;
		this.from_port = from_port;
		this.to_port = to_port;
		this.total_bits = total_bits;
		this.remaining_bits = total_bits;
		this.flow = flow;
		this.flow.attach_monitor(this);
		this.bw = this.oldbw = flow.get_bw();
		this.monitors = new LinkedList();
		if(DEBUG) s.debug(this,"created: size="+total_bits);
	}
}
