package flowsim.core;
import flowsim.util.CompositeIterator;

import java.util.Enumeration;

import com.sun.java.util.collections.*;

final class FairLMFlowsList extends LinkedList
{
	public void add(Flow f) {
		ListIterator it = listIterator();
		while(it.hasNext()) {
			Flow i = (Flow) it.next();
			if(f.get_bw()>i.get_bw()) break;
		}
		it.add(f);	// warning: semantics will change in JDK 1.2 (it
				// now inserts before, and will insert after)...
	}
}

final public class FairLinkManager extends LinkManagerImpl implements FlowMonitor
{
	final static boolean DEBUG = false;
	final static boolean CHECK = false;
	final static float   RESERVABLE = (float) 0.95;
	final Simulator s;
	final FairLMFlowsList be_flows;
	final FairLMFlowsList nbe_flows;
	final int reservable_bw;
	int nbe_bw; // bw of non-be flows on link

	public static LinkManagerFactory factory(final Simulator s) {
		return new LinkManagerFactory() {
			public LinkManager create(Link link) {
				return new FairLinkManager(s,link);
			}
		};
	}

	public int get_reservable_bw() {
		return reservable_bw;
	}

	public Iterator flows() {
		//s.debug(this,"be_flows="+be_flows.size()+", nbe_flows="+nbe_flows.size());
		return new CompositeIterator(be_flows.iterator(), nbe_flows.iterator());
	}

	boolean is_be(Flow flow) {
		if(flow.get_ds() == FlowImpl.DS_BEST_EFFORT) return true;
		return false;
	}

	public long add(Flow flow) {
		if(is_be(flow)) {
			if(DEBUG) s.debug(this,"adding best-effort flow.");

			remove_be_flows();
			be_flows.add(flow);
			readd_be_flows();
		}
		else {
			if(DEBUG) s.debug(this,"adding non-be flow.");

			int bw = flow.get_desired_bw();

			if(nbe_bw+bw>reservable_bw) {
				if(DEBUG) s.debug(this, "blocking flow...");
				flow.block();
				return -1;
			}

			remove_be_flows();

			flow.set_bw(bw,bw);

			if(link.add_flow(bw) == false) {
				s.error(this,"link blocked non-be flow.");
				return -1;
			}

			nbe_bw += bw;

			if(CHECK) {
				if(link.get_rbw()!=link.get_bw()-nbe_bw) {
					s.error(this,"CHECK FAILURE: "+
						link.get_rbw()+"!="+
						(link.get_bw()-nbe_bw));
				}
			}

			nbe_flows.add(flow);

			readd_be_flows();
		}

		flow.attach_monitor(this);

		return super.add(flow); // monitors
	}

	void remove_be_flows() {
		// remove every BE flow (counting them).
		Iterator it = be_flows.iterator();
		while(it.hasNext()) {
			Flow f = (Flow) it.next();
			if(DEBUG) s.debug(this,"removing (tmp) flow (bw="+f.get_bw()+", flow="+f+").");
			link.remove_flow(f.get_bw());
		}
	}

	void readd_be_flows() {
		int be_flows_nr = be_flows.size();

		if(be_flows_nr<=0) return;

		int rbw =(link.get_bw()-nbe_bw);


		Iterator it = be_flows.iterator();
		while(it.hasNext()) {
			Flow f = (Flow) it.next();

			int bw = rbw/be_flows_nr;
			int fbw;
			if(f.get_desired_bw()<bw) fbw = f.get_desired_bw();
			else                  fbw = bw;

			if(DEBUG) s.debug(this,"adding flow (bw="+bw+", desired="+
				f.get_desired_bw()+", fbw="+fbw+").");

			if(!link.add_flow(fbw)) {
				s.error(this,"Couldn't add flow to link!"+
				" link.rbw="+link.get_rbw());
				return;
			}
			f.set_bw(bw, fbw);
			rbw -= fbw;
			be_flows_nr--;
		}
	}

	void adjust_be_flows() {
		remove_be_flows();
		readd_be_flows();
	}

	public int get_rbw_notbe() {
		return link.get_bw()-nbe_bw;
	}

	/** Explicitely removes a flow. A flow is automatically removed when
	 *  it is stopped or blocked.
	 */
	public void remove(Flow flow) {
		if(DEBUG) s.debug(this,"removing flow "+flow);

		int bw = flow.get_bw();

		if(is_be(flow)) {
			if(!be_flows.remove(flow)) {
				if(DEBUG) s.debug(this,"Flow "+flow+" already removed.");
				return;
			}
		}
		else {
			if(!nbe_flows.remove(flow)) {
				if(DEBUG) s.debug(this, "Flow "+flow+" already removed.");
				return;
			}
			nbe_bw -= bw;
			if(nbe_bw<0) {
				s.error(this,"SimpleLinkManager: nbe_bw<0");
				return;
			}
		}

		link.remove_flow(bw);

		adjust_be_flows();

		super.remove(flow); // monitors
	}

	public void flow_started(Flow f) {}

	public void flow_stopped(Flow f) {
		remove(f);
	}

	public void flow_blocked(Flow f) {
		remove(f);
	}

	public void flow_changed_bw(Flow f, int was_bw) {}

	public void flow_changed_desired_bw(Flow f, int was_bw) {
		if(DEBUG) s.debug(this,"Flow "+f+" changed desired bandwidth to "+f.get_desired_bw());
		adjust_be_flows();
	}
	public void flow_changed_available_bw(Flow f, int was_bw) {}

	final public String toString() {
		return "FairLinkMan("+ link.get_from()+"-"+
				link.get_to()+")";
	}

	public void print_debug_info() {
		System.out.println("SimpleLinkManager: nbe_bw="+nbe_bw);
		int nbe_bw; // bw of non-be flows on link
	}

	public FairLinkManager(Simulator s, Link link) {
		super(link);
		this.s = s;
		be_flows = new FairLMFlowsList();
		nbe_flows = new FairLMFlowsList();
		reservable_bw = (int) (link.get_bw()*RESERVABLE);
	}
}
