package flowsim.core;

/**
 * A flow simulates a stream of data packets in the network. Flow objects
 * have a source node (from), bandwidth (bw), flow-id (flowid) and
 * differentiated-services-id (ds). They can be started, stopped, blocked and
 * forwarded to an interface.
 *
 * @see Node
 */ 
public interface Flow {
	/**
	 * Get source node.
	 * @return source node.
	 */
	Node get_from();

	/**
	 * Get source service. Optional, can return null.
	 * @return source service or null if source wasn't a service (or
	 * didn't tell so)
	 */
	Service get_from_service();

	/**
	 * Set source service.
	 */
	void set_from_service(Service from);

	/**
	 * Get bandwidth used/required by this flow.
	 * @return bandwidth in bits/second.
	 */
	int get_bw();

	/**
	 * Get wanted bandwidth of this flow.
	 * @return bandwidth in bits/second.
	 */
	int get_desired_bw();
	
	/**
	 * Get maximal bandwidth which will be given to this flow, if it
	 * "desired" maximal bandwidth.
	 * @return bandwidth in bits/second.
	 */
	int get_available_bw();
	
	/**
	 * Set  available and used bandwidth by this flow. This method should
	 * be only called by LinkManagers. Use set_desired_bw to set the
	 * wanted bandwidth of the flow, the final decision belongs the
	 * LinkManager...
	 * @param bw bandwidth in bits/second.
	 */
	void set_bw(int available_bw, int bw);

	/**
	 * Set wanted bandwidth of this flow. This value will be used by the
	 * LinkManager to determine the real allocated bandwidth, which will
	 * be set by him with 'set_bw'.
	 * @param bw bandwidth in bits/second.
	 * @see LinkManager
	 */
	void set_desired_bw(int bw);

	/**
	 * Set flow-id of this flow. The flow-id can be used for example to
	 * choose a color in a nam-trace or to recognise a certain type of
	 * flows.
	 */
	int get_flowid();

	/**
	 * Get flow-id of this flow.
	 */
	void set_flowid(int flowid);

	/**
	 * Get differentiated-services 'byte' of this flow. This field, also
	 * known as 'type-of-service' (TOS), is used in diffserv networks.  It
	 * is normally a byte, but is in flowsim for the sake of commodity an
	 * 'int'.
	 */
	int get_ds();

	/**
	 * Set differentiated-services 'byte' of this flow.
	 */
	void set_ds(int ds);

	/**
	 * Start the flow of data.
	 */
	void start();

	/**
	 * Stop the flow.
	 */
	void stop();

	/**
	 * Block the flow. Sets the 'blocked' status and stops the flow.
	 */
	void block();

	/**
	 * Is this flow still active (i.e. are packets still sent?)
	 */
	boolean is_running();

	/**
	 * Was this flow blocked?
	 */
	boolean is_blocked();

	/**
	 * Utility of this flow.
	 * @return utility in the range [0,1].
	 */
	double utility();

	/**
	 * Add a FlowMonitor to this flow. The FlowMonitor will be notified
	 * when the flow starts, stops, blocks and changes bandwidth.
	 */
	void attach_monitor(FlowMonitor monitor);

	/**
	 * Remove a FlowMonitor from this flow.
	 */
	 void detach_monitor(FlowMonitor monitor);
}
