package flowsim.core;

/**
 * A link object simulates the physical behaviour of a real-link. It does
 * have a total bandwidth, which can be used by flows which circulate on it.
 * The unused bandwidth is called 'residual bandwidth'. A link is in flowsim
 * like a bucket which can be filled up-to it's capacity.
 * The delay (for the added packet, maximal and minimal) can be also asked
 * from the link object.<P>
 * The packet queuing implementation could be simulated in flowsim by
 * modelling the delay with a function of the residual bandwidth.
 *
 * @author David Schweikert <schweikert@acm.org>
 * @see flowsim.core.FixedDelayLink
 */
public interface Link
{
	/**
	 * Get the source node.
	 */
	Node get_from();

	/**
	 * Get the destination node.
	 */
	Node get_to();

	/**
	 * Get the total bandwidth of the link.
	 */
	int  get_bw();

	/**
	 * Get the residual bandwidth of the link.
	 */
	int  get_rbw();

	/**
	 * Get the (probable) delay which would have a flow sent now.
	 */
	int get_delay(); /* in microseconds (us) */

	/**
	 * Get the minimal delay that a flow can take on this link.
	 */
	int get_min_delay();

	/**
	 * Get the maximal delay that a flow can take on this link.
	 */
	int get_max_delay();

	/**
	 * Add a flow on the link.
	 */
	boolean add_flow(int fbw);

	/**
	 * Remove a flow from the link.
	 */
	void remove_flow(int fbw);

	/**
	 * Attach a LinkMonitor on the link.
	 */
	void attach_monitor(LinkMonitor lmo);

	/**
	 * Get attached monitor.
	 */
	LinkMonitor get_monitor();
}
