package flowsim.core;

import com.sun.java.util.collections.Iterator;

/**
 * Admission Control/Queuing for a Link. A LinkManager decides which flows are
 * blocked and which are accepted. It does also modify the bandwidth of the
 * flows on the link. LinkManagers are used to model network interfaces on a
 * node.
 *
 * @see Link
 * @see Node
 */
public interface LinkManager
{
	/**
	 * Get the flows managed by this LinkManager.
	 */
	Iterator flows();

	/**
	 * Get node on the other side of the link.
	 * @return peer node.
	 */
	Node get_peer();

	/**
	 * Get managed link.
 	 */
	Link get_link();

	/**
	 * Get residual bandwith without counting best-effort.
	 * @return residual bandwith in bits/second.
	 */
	int get_rbw_notbe();

	/**
	 * Get reservable bandwidth (bandwidth useable by non-be flows).
	 */
	int get_reservable_bw();

	/**
	 * Add new flow.
	 * @return delay in micro-seconds.
	 */
	long add(Flow flow);

	/**
	 * Called when the bandwidth of a flow is changed (usually when it is
	 * made smaller by other linkmanagers.
	 */
	void bw_changed(Flow flow);

	/**
	 * Remove flow.
	 */
	void remove(Flow flow);

	/**
	 * Attach a LinkManagerMonitor, which will then be notified when a
	 * flow is added, removed or when it's bandwidth was changed.
	 * Note that it should be possible to attach multiple monitors.
	 */
	void attach_monitor(LinkManagerMonitor lmm);

	/**
	 * Detach a LinkManagerMonitor.
	 */
	void detach_monitor(LinkManagerMonitor lmm);
}
