package flowsim.core;

/**
 * A node consists of an id/name, interfaces and services on its ports.
 * Every interface is a LinkManager instance.
 *
 * @see LinkManager
 * @see Service
 */
public interface Node
{
	/**
	 * Get id of this node. This number should be unique among every
	 * created nodes.
	 */
	int    get_id();

	/**
	 * Get name of this node. Used for print/debug/display purproses.
	 */
	String get_name();

	/**
	 * Get maximum number of interfaces which can be added to this node.
	 */
	int  get_max_ifaces();

	/**
	 * Get number of connected (!=null) interfaces.
	 */
	int  get_ifaces_count();

	/**
	 * Set interface number n to a LinkManager.
	 * @param f interface instance, null if none.
	 */
	void set_iface(int n, LinkManager f);

	/**
	 * Add an interface on the first available iface number.
	 * @return iface number.
	 */
	int add_iface(LinkManager f);

	/**
	 * Get interface (modelled by a LinkManager) number i.
	 */
	LinkManager get_iface(int i);

	/**
	 * Get iface-number which goes to a node.
	 * @return iface-number or -1 if node isn't a peer.
	 */
	int get_iface(Node n);

	/**
	 * Get Link instance of this interface.
	 * This is like get_iface(n).get_link().
	 */
	Link get_link(int n);
	
	/**
	 * Get peer-node of this interface.
	 * This is like get_link(n).get_peer().
	 */
	Node get_peer(int n);

	/**
	 * Get service (port listener).
	 * @param port port of the service.
	 */
	Service get_service(int port);

	/**
	 * Set service (port listener).
	 * @param service Service instance.
	 * @param port Port number (first is 0).
	 */
	void set_service(Service service, int port);

	/**
	 * Add monitor to this node, which will be notified when the ifaces
	 * change, ...
	 * @param nm the monitor
	 */
	public void attach_monitor(NodeMonitor nm);

	/**
	 * Remove monitor from this node.
	 * @param nm the monitor
	 */
	public void remove_monitor(NodeMonitor nm);

}
