package flowsim.core;

import java.util.Enumeration;

import com.sun.java.util.collections.Collection;
import com.sun.java.util.collections.Collections;
import com.sun.java.util.collections.Comparator;
import com.sun.java.util.collections.Iterator;
import com.sun.java.util.collections.TreeSet;

final class SimpleLMComparator implements Comparator {

	final int cmp(final int a, final int b) {
		if(a>b) return 1;
		else if(a<b) return -1;
		return 0;
	}

	public final int compare(Object _a, Object _b) {
		Flow a = (Flow) _a;
		Flow b = (Flow) _b;
		
		final int cmp1 = cmp(a.get_ds(),b.get_ds());
		if(cmp1!=0) return cmp1;

		final int cmp2 = cmp(a.get_desired_bw(),b.get_desired_bw());
		if(cmp2!=0) return cmp2;

		return cmp(a.hashCode(),b.hashCode());
	}
}

final class SimpleLMFlowsList extends TreeSet implements Cloneable {
	Iterator be_flows() {
		final int ds_be = FlowImpl.DS_BEST_EFFORT;
		return subSet(new FlowImpl((Node) null,0,0,ds_be),
			new FlowImpl((Node) null,0,0,ds_be+1)).iterator();
	}
	Iterator non_be_flows() {
		final int ds_be = FlowImpl.DS_BEST_EFFORT;
		return subSet(new FlowImpl((Node) null,0,0,ds_be+1),
			new FlowImpl((Node) null,0,0,Integer.MAX_VALUE)).iterator();
	}
	SimpleLMFlowsList() {
		super(new SimpleLMComparator());
	}
}

/**
 * Simple link manager which does recognise two type of classes: best-effort
 * and non-best-effort. Best-effort are considered every flow with DS
 * FlowImpl.DS_BEST_EFFORT. The rest of bandwidth not used by non-best-effort
 * flows is equally divided among all best-effort flows (bandwidth is forced
 * to that value). If a non-best-effort flow is added for which there isn't no
 * space, it is blocked.
 * @see FairLinkManager
 */

final public class SimpleLinkManager extends LinkManagerImpl implements FlowMonitor
{
	final static boolean DEBUG = true;
	final static boolean CHECK = true;
	final Simulator s;
	final SimpleLMFlowsList flows;
	int non_be_bw; // bw of non-be flows on link

	public static LinkManagerFactory factory(final Simulator s) {
		return new LinkManagerFactory() {
			public LinkManager create(Link link) {
				return new SimpleLinkManager(s,link);
			}
		};
	}

	public Iterator flows() {
		return flows.iterator();
	}

	public int get_reservable_bw() {
		return link.get_bw();
	}

	public long add(Flow flow) {
		if(flow.get_ds() == FlowImpl.DS_BEST_EFFORT) {
			if(DEBUG) s.debug(this,"adding best-effort flow.");

			flows.add(flow);
			adjust_be_flows();
		}
		else {
			if(DEBUG) s.debug(this,"adding non-be flow.");

			int be_flows = remove_be_flows();

			int bw = flow.get_bw();
			if(link.add_flow(bw) == false) {
				if(true) s.debug(this,"blocking non-be flow "+
					"(link: bw="+link.get_bw()+
					", rbw="+link.get_rbw()+")");
				flow.block();
				readd_be_flows(be_flows);
				return -1;
			}
			non_be_bw += bw;
			if(non_be_bw>link.get_bw()) {
				s.error(this,"non_be_bw>link.bw");
			}
			flows.add(flow);

			readd_be_flows(be_flows);
		}

		flow.attach_monitor(this);

		return super.add(flow); // monitors
	}

	int remove_be_flows() {
		// remove every BE flow (counting them).
		int be_flows=0;
		Iterator it = flows.be_flows();
		while(it.hasNext()) {
			Flow f = (Flow) it.next();
			if(f.get_ds()!=FlowImpl.DS_BEST_EFFORT) {
				s.error(this,"flows.search_type behaving incorrectly!");
				continue;
			}
			if(DEBUG) s.debug(this,"removing (tmp) flow (bw="+f.get_bw()+", flow="+f+").");
			link.remove_flow(f.get_bw());
			// f.set_bw(0);
			be_flows++;
		}

		if(CHECK) {
			if(non_be_bw != (link.get_bw()-link.get_rbw())) {
				s.error(this,
						"Error: non_be_bw="+non_be_bw+
						" != (link.bw-link.rbw)="+(link.get_bw()-link.get_rbw()));

				System.err.println();
				it = flows.iterator();
				while(it.hasNext()) {
					Flow f = (Flow) it.next();
					if(f.get_ds()==FlowImpl.DS_BEST_EFFORT) {
						System.err.print("BE\t");
					}
					else {
						System.err.print("\t");
					}
					System.err.println(""+f.get_bw());
				}
				System.err.println();
			}
		}
		
		return be_flows;
	}

	void readd_be_flows(int be_flows) {
		if(be_flows>0) {
			// re-add them
			int bw = (link.get_bw()-non_be_bw)/be_flows;
			Iterator it = flows.be_flows();
			while(it.hasNext()) {
				Flow f = (Flow) it.next();
				if(f.get_ds()!=FlowImpl.DS_BEST_EFFORT) {
					s.error(this,"flows.search_type behaving incorrectly!");
					continue;
				}
				if(DEBUG) s.debug(this,"adding flow (bw="+bw+", flow="+f+").");
				f.set_bw(bw,bw);
				if(link.add_flow(bw) == false ) {
					if(DEBUG) s.debug(this,"be-flow blocked.");
					s.error(this,"be-flow blocked.");
					f.block();
					flows.remove(f);
				}
			}
		}
	}

	void adjust_be_flows() {
		int be_flows=remove_be_flows();
		readd_be_flows(be_flows);
	}

	public int get_rbw_notbe() {
		return link.get_bw()-non_be_bw;
	}

	/** Explicitely removes a flow. A flow is automatically removed when
	 *  it is stopped or blocked.
	 */
	public void remove(Flow flow) {
		if(DEBUG) s.debug(this,"removing flow "+flow);
		if(!flows.remove(flow)) {
			if(DEBUG) s.debug(this,"Flow "+flow+" already removed.");
			return;
		}

		int bw = flow.get_bw();
		link.remove_flow(bw);
		if(flow.get_ds()!=FlowImpl.DS_BEST_EFFORT) {
			non_be_bw -= bw;
			if(non_be_bw<0) {
				s.error(this,"SimpleLinkManager: non_be_bw<0");
			}
		}

		adjust_be_flows();

		super.remove(flow); // monitors
	}

	public void flow_started(Flow f) {}

	public void flow_stopped(Flow f) {
		remove(f);
	}

	public void flow_blocked(Flow f) {
		remove(f);
	}

	public void flow_changed_bw(Flow f, int was_bw) {}
	public void flow_changed_desired_bw(Flow f, int was_bw) {}
	public void flow_changed_available_bw(Flow f, int was_bw) {}

	final public String toString() {
		return "SimpleLinkMan("+ link.get_from()+"-"+
				link.get_to()+")";
	}

	public void print_debug_info() {
		System.out.println("SimpleLinkManager: non_be_bw="+non_be_bw);
		int non_be_bw; // bw of non-be flows on link
	}

	public SimpleLinkManager(Simulator s, Link link) {
		super(link);
		this.s = s;
		flows = new SimpleLMFlowsList();
	}
}
