package flowsim.core;

import flowsim.util.StringFormatter;

import java.util.Random;

import java.io.*;

/** 
 * A Simulator object does contain all the services which are needed
 * globally by every element such as the scheduler.
 */
public final class Simulator
{
	static final int     RUN_MAX_TIME=50000;
	static final int     PROGRESS_TICS=40;
	final Scheduler scheduler;
	final Logger logger;
	int next_node_id=1;
	boolean verbose_run;

	/**
	 * Get the current time of the simulation in microseconds.
	 */
	public long now() {
		return scheduler.now();
	}

	class ProgressEvent implements Event {
		public void trigger() {
			System.err.print("."); System.err.flush();
		}
	}
	void show_progress(long time) {
		long dt = time/PROGRESS_TICS;
		for(int i=1; i<=PROGRESS_TICS; i++) {
			schedule(dt*i, new ProgressEvent());
		}
	}

	/**
	 * Runs the simulation.
	 * @param how_long duration in microseconds.
	 */
	public void run(final long how_long) {
		if(how_long>RUN_MAX_TIME) {
			run(RUN_MAX_TIME);
			run(how_long-RUN_MAX_TIME);
			return;
		}

		if(verbose_run) {
			String start = StringFormatter.right_justified(""+
				now(),10,true);
			String stop = StringFormatter.right_justified(""+
				(now()+how_long),10,true);
			System.err.print("Simulator run: "+start+" - "+stop+" : ");
			System.err.flush();
			show_progress(how_long);
		}

		scheduler.run(how_long);
		
		if(verbose_run) {
			System.err.println();
		}
	}

	/**
	 * Schedules an event at a time relative to now.
	 */
	public EventQueueElement schedule(final long when, final Event e) {
		return scheduler.schedule(when,e);
	}

	/**
	 * Schedules an event at absolute time.
	 */
	public EventQueueElement schedule_absolute(final long when, final Event e) {
		return scheduler.schedule_absolute(when,e);
	}



	/**
	 * Writes an error string to the log.
	 */
	public void error(final Object who, final String str) {
		logger.error(who,str);
	}

	/**
	 * Writes a debug to the log.
	 */
	public void debug(final Object who, final String str) {
		logger.debug(who,str);
	}

	public void set_verbose(boolean v) {
		verbose_run = v;
	}

	public Simulator() {
		scheduler = new Scheduler(this);
		logger = new Logger(this,System.err,System.err);
		verbose_run=false;

		/*
		try {
			FileOutputStream fos = new FileOutputStream("debug.log");
			BufferedOutputStream bos = new BufferedOutputStream(fos);
			PrintStream ps = new PrintStream(bos);

			logger = new Logger(this,System.err,ps);
		}
		catch(IOException ioe) {}
		*/
	}
}
