package flowsim.core;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Random;

public class Topology
{
	final Simulator s;
	final String logs_dir;
	final int max_neighbours;
	final int max_services;

	Node[] n;
	Vector links;
	Vector linkmanagers;
	LinkMonitor total_bw_monitor;

	public int nodes() { return n.length; }
	public Node node(int nnr) { return n[nnr]; }

	public NodeChooser get_random_node_chooser(final Random r) {
		return new NodeChooser() {
			public Node get() {
				int node_nr = (int) (n.length*r.nextDouble());
				return n[node_nr];
			}
		};

	}

	public void add_services(ServiceFactory rf, int port) {
		for(int i=0; i<n.length; i++) {
			rf.create(n[i],port);
		}
	}

	public void add_link(int node_a, int node_b, LinkFactory lf, LinkManagerFactory lmf)
	{
		final Node a = n[node_a];
		final Node b = n[node_b];

		// Create link
		Link l = lf.create(a,b);

		// Create logs (link monitors)
		mklinklog(l,a,b);

		// Create link managers
		LinkManager lm = lmf.create(l);
		a.add_iface(lm);

		// DEBUG (REMOVE) 
		/*
		if(a.get_id()==3) {
			LinkManagerMonitor lmm = new LinkManagerLogger();
			lm.attach_monitor(lmm);
		}
		*/

		// Add link to the 'links' vector
		links.addElement(l);
	}

	PrintWriter mklinklog(final Link l, final Node a, final Node b) {
		if(logs_dir==null) return null;

		try {
			String filename = new String(logs_dir+"/"+"bw_"+
				a+"-"+b+".log");
			FileOutputStream f = new FileOutputStream(filename);
			PrintWriter p = new PrintWriter(f,true);
			LinkBandwidthLogger m = new LinkBandwidthLogger(s,l,p);
			l.attach_monitor(m);
			m.attach_monitor(total_bw_monitor);
		}
		catch(IOException e) {
			System.out.println("IOException");
		}
		return null;
	}

	public void add_duplex_link(int node_a, int node_b,
		LinkFactory lf, LinkManagerFactory lmf)
	{
		add_link(node_a,node_b,lf,lmf);
		add_link(node_b,node_a,lf,lmf);
	}

	public void remove_link(int node_a, int node_b) {
		final Node a = n[node_a];
		final Node b = n[node_b];

		int ifaces = a.get_max_ifaces();
		for(int i=0; i<ifaces; i++) {
			if(a.get_peer(i)==b) {
				a.set_iface(i,null);
				break;
			}
		}
	}

	public void remove_duplex_link(int node_a, int node_b) {
		remove_link(node_a, node_b);
		remove_link(node_b, node_b);
	}

	public void flush_logs() {
		Enumeration e = links.elements();
		while(e.hasMoreElements()) {
			Link l = (Link) e.nextElement();
			//FIXME!
			if(l.get_monitor()!=null) l.get_monitor().flush();
		}
	}

	void init_total_bw() {
		if(logs_dir==null) return;
		try {
			String filename = new String(logs_dir+"/total_bw.log");
			FileOutputStream f = new FileOutputStream(filename);
			PrintWriter pw = new PrintWriter(f,true);
			total_bw_monitor = new TotalBandwidthMonitor(s, pw);
		}
		catch(IOException e) {
			System.out.println("IOException");
		}
	}

	public void make_nam_log(String filename)
	{
		make_nam_log(filename,LinkManagerNamLogger.DEFAULT_CYCLE_TIME,false);
	}

	public void make_nam_log(String filename, long cycle_time,
		boolean flowid_aggregates)
	{
		PrintWriter p;
		try {
			FileOutputStream f = new FileOutputStream(filename);
			p = new PrintWriter(f,true);
		}
		catch(IOException e) {
			System.out.println("IOException");
			return;
		}
		
		p.println("c -t * -i 0 -n red");
		p.println("c -t * -i 1 -n orange");
		p.println("c -t * -i 2 -n yellow");
		p.println("c -t * -i 3 -n green");
		p.println("c -t * -i 4 -n blue");
		p.println("c -t * -i 5 -n magenta");
		p.println("c -t * -i 6 -n violet");
		p.println("c -t * -i 7 -n black");
		p.println("c -t * -i 8 -n white");
		p.println("c -t * -i 9 -n grey");

		for(int node=0; node<n.length; node++) {
			p.println("n -t * -s "+n[node]+" -S UP -v circle");
		}
		for(int node=0; node<n.length; node++) {
			for(int iface=0; iface<max_neighbours; iface++) {
				LinkManager lm = n[node].get_iface(iface);
				if(lm!=null && lm.get_peer().get_id()>n[node].get_id()) {
					p.println("l -t * -s "+n[node]+
					" -d "+lm.get_peer()+" -S UP "+
					" -r "+lm.get_link().get_bw()+"b"+
					" -D "+lm.get_link().get_delay()+"us");
				}
			}
		}
		for(int node=0; node<n.length; node++) {
			for(int iface=0; iface<max_neighbours; iface++) {
				LinkManager lm = n[node].get_iface(iface);
				if(lm!=null) {
					//System.out.println("Adding LinkManagerNamLogger...");
					LinkManagerNamLogger m =
						new LinkManagerNamLogger(s,n[node],lm,
							p,cycle_time,flowid_aggregates);
					lm.attach_monitor(m);
				}
			}
		}
	}

	public Topology(Simulator s, int nodes,
		int max_neighbours, int max_services, String logs_dir)
	{
		this.s = s;
		this.max_neighbours = max_neighbours;
		this.max_services = max_services;
		this.logs_dir = logs_dir;

		links = new Vector();
		linkmanagers = new Vector();

		n = new NodeImpl[nodes];
		for(int i=0; i<nodes; i++) {
			String name = Integer.toString(i);
			n[i] = new NodeImpl(s,name,max_neighbours,max_services);
		}

		init_total_bw();
	}
}
