package flowsim.core;

// A unicast flow from a to d, through b and c is so constructed:

//           a -------- b ---------- c --------- d
// Flow1:    a --------------------------------> d
// Flow2:               b ---------------------> d
// Flow3:                            c --------> d

// Flow1 is parent of Flow2 and Flow2 is child of Flow1...

public class UnicastFlowImpl extends RoutedFlowImpl implements UnicastFlow
{
	static final boolean DEBUG=false;
	Node to;
	LinkManager iface;
	UnicastFlow parent;
	UnicastFlow child;
	int parent_bw;

	class SetChildBwEvent implements Event {
		int pbw;
		public void trigger() { child.set_parent_bw(pbw); }
		SetChildBwEvent(int pbw) { this.pbw = pbw; }
	}
	
	protected void set_parent(UnicastFlow parent) {
		this.parent = parent;
	}
	
	protected UnicastFlow mkchild(Node child_node) {
		if(DEBUG) s.debug(this,"making child on node "+child_node);
		UnicastFlowImpl f = new UnicastFlowImpl(s,child_node,to,bw,flow_id,ds,
			router_port,this);
		return f;
	}

	/**
	 * Set the child to a clone of this flow. This is used for for example
	 * in special cases of routing, where another object has a reference
	 * to the flow, but you want to modify it (for example the ds byte).
	 * This is very special and is normally not used. Note that you have
	 * to start() the flow after that.
	 * @return Child flow originating on the same node
	 */
	public UnicastFlow add_child_same_node() {
		child = mkchild(from);
		return child;
	}

	public LinkManager get_iface() {
		return iface;
	}

	public void next_node(Node child_node) {
		// verify if it was blocked or stopped in the meantime
		if(blocked || !running) { // FIXME 3.6.99
			if(DEBUG) s.debug(this,"Blocked in the meantime...");
			return;
		}

		if(DEBUG) s.debug(this,"arrived at node: "+child_node);

		// generate child flow on next node
		if(child_node != to) {
			child = mkchild(child_node);
			child.start();
		}
	}

	public final Node get_to()  { return to; }

	public final UnicastFlow get_parent() {
		return parent;
	}

	public final UnicastFlow get_child() { return child; }

	public void forward(LinkManager lm) {
		super.forward(lm);
		if(DEBUG) s.debug(this,"forwarded to lm "+lm);
		iface = lm;
	}

	class StopChildEvent implements Event {
		public void trigger() {
			child.stop();
		}
	}

	void stop_child() {
		if(iface!=null) {
			s.schedule(iface.get_link().get_delay(), new StopChildEvent());
		}
		else {
			child.stop();
		}
	}

	public void stop() {
		//s.debug(this,"stopping");

		if(running) {
			if(child!=null) stop_child();
		}

		super.stop();
	}

	public void block() {
		if(parent!=null) {
			if(DEBUG) s.debug(this,"Telling my parent that I was blocked...");
			parent.block();
		}

		super.block();
	}

	public void set_parent_bw(int bw) {
		if(parent_bw==bw) return;
		if(DEBUG) s.debug(this,"set_parent_bw("+bw+"), old="+parent_bw);
		parent_bw = bw;
		set_desired_bw(bw);
	}

	public void set_bw(int abw, int bw) {
		int old_bw = this.bw;
		
		if(DEBUG) s.debug(this,"set_bw("+abw+", "+bw+")");

		super.set_bw(abw, bw);

		if(old_bw != bw && child!=null) {
			if(iface!=null) {
				if(DEBUG) s.debug(this,"telling child...");
				s.schedule(iface.get_link().get_delay(), new SetChildBwEvent(bw));
			}
			else {
				if(DEBUG) s.debug(this,"telling local child...");
				child.set_parent_bw(bw);
			}
		}
	}

	public String toString() {
		//return "UnicastFlow("+from+","+to+","+ds+")";
		return "UnicastFlow("+
			Integer.toHexString(this.hashCode())+":"+
			from+","+to+","+ds+")";
	}

	UnicastFlowImpl(Simulator s, Node from, Node to, int desired_bw, int flow,
		int ds, int rport, UnicastFlow parent)
	{
		super(s,from,desired_bw,flow,ds,rport);
		this.s  = s;
		this.to = to;
		this.desired_bw = this.parent_bw = this.bw = desired_bw;
		this.parent=parent;

		if(DEBUG) s.debug(this,"instantiated: desired_bw = "+desired_bw);
	}

	public UnicastFlowImpl(Simulator s, Node from, Node to, int desired_bw, int flow,
		int ds, int rport)
	{
		this(s,from,to,desired_bw,flow,ds,rport,null);
	}
}
