package flowsim.dv;

import flowsim.core.*;

import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;

public class DVRouterImpl extends RouterImpl implements DVRouter
{
	static final boolean DEBUG=false;
	public static final int INFINITY=1000;

	int      ifaces;
	DVTable[]  ntables;
	Dictionary nnodes;
	int[]    nodes_count;
	DVTable    cost_table;
	Dictionary iface_table;

	class IfaceEntry
	{
		int iface=-1;
	}

	public static ServiceFactory factory(final Simulator s) {
		return new ServiceFactory() {
			public Service create(Node node, int route) {
				return new DVRouterImpl(s,node,route);
			}
		};
	}

	public void receive_update(Node from, DVTable t) {
		if(DEBUG) s.debug(this,"received update from "+from+
			" (time = "+s.now()+")");
		int i = node.get_iface(from);
		if(i>=0) {
			set(i,t);
			run();
		}
	}

	final void set_iface(Node n, int iface) {
		IfaceEntry e = (IfaceEntry) iface_table.get(n);
		e.iface = iface;
	}

	final void set(int iface, DVTable t) {
		ntables[iface] = t;
		if(t==null) return;

		// see if keys (nodes) in table contains all the nodes in t.
		if(nodes_count[iface]<t.size()) {
			for(Enumeration e = t.nodes(); e.hasMoreElements();) {
				Node n = (Node) e.nextElement();
				if(iface_table.get(n)==null) {
					cost_table.put(n,INFINITY);
					iface_table.put(n,new IfaceEntry());
				}
			}
			nodes_count[iface]=t.size();
		}
	}

	final void init_neighbour(int iface) {
		DVTable t;
		LinkManager lm = node.get_iface(iface);
		if(lm==null) {
			// iface is removed
			nodes_count[iface]=0;
			t = null;
		}
		else {
			// iface is added or changed
			t = new DVTableImpl();
			t.put(lm.get_peer(),0);
		}

		set(iface,t);
		
		// send update to him
		// send_update(iface);
	}

	protected int cost_threshold() {
		return 1; // no threshold.
	}

	protected int link_cost(int iface) {
		return 1;
	}

	final int abs(int x) { if(x<0) return -x; return x; }
	final boolean compute() { // true if table changes
		if(DEBUG) s.debug(this,"recomputing routing table.");

		boolean changed = false;

		for(Enumeration e = cost_table.nodes(); e.hasMoreElements();) {
			Node n = (Node) e.nextElement();

			int cost = INFINITY;
			int iface=-1;

			if(n==node) {
				// ok, it's us. cost = 0;
				cost = 0;
			}
			else {
				// find minimum
				for(int i=0; i<ifaces; i++) {
					int c;
					if(ntables[i]==null) c = INFINITY;
					else c = ntables[i].get(n) + link_cost(i);
					if(c<cost) {
						cost = c;
						iface = i;
					}
				}
			}

			// s.debug(this,"cost for node "+n.get_name()+": "+cost+" (iface="+iface+")");

			// is old value the same?
			int oldcost = cost_table.get(n);
			if(abs(cost-oldcost) >= cost_threshold()) {
				if(DEBUG) s.debug(this,"oldcost="+oldcost+", cost="+cost);
				changed = true;
				cost_table.put(n,cost);
				set_iface(n,iface);
			}
		}
		return changed;
	}

	final void send_update(int iface) {
		LinkManager lm = node.get_iface(iface);
		if(lm!=null) {
			int count      = iface_table.size();
			Link l         = lm.get_link();
			DVTable table  = (DVTable) cost_table.clone();
			DVRouter router = (DVRouter) l.get_to().get_service(get_port());

			Message msg = new DVUpdateMessage(node, router, table, count);
			Packet packet = new NextHopPacket(s,node,lm,msg,0);

			packet.start();
		}
	}

	final void send_updates() {
		for(int i=0; i<ifaces; i++) {
			send_update(i);
		}
	}

	protected void run() {
		if(DEBUG) s.debug(this,"run()");
		if(compute()) {
			if(DEBUG) s.debug(this,"something changed: sending updates.");
			send_updates();
		}
	}

	final public void iface_changed(Node n, int i) {
		super.iface_changed(n, i);

		if(DEBUG) s.debug(this,"neighbour changed (iface nr. "+i+").");
		init_neighbour(i);
		run();
	}

	public final String toString() {
		return "DVRouter("+node+")";
	}

	public void debug_table() {
		System.out.print("\t");
		for(int i=0; i<ifaces; i++) {
			LinkManager lm = node.get_iface(i);
			if(lm != null) {
				System.out.print(lm.get_peer()+"\t");
			}
		}
		System.out.println("cost\tiface");
		for(Enumeration e = cost_table.nodes(); e.hasMoreElements();) {
			Node n = (Node) e.nextElement();

			System.out.print(n+"\t");
			
			for(int i=0; i<ifaces; i++) {
				LinkManager lm = node.get_iface(i);
				if(lm != null) {
					int cost;
					if(ntables[i]==null) cost = INFINITY;
					else cost = ntables[i].get(n) + link_cost(i);
					System.out.print(cost+"\t");
				}
			}

			// cost
			System.out.print(cost_table.get(n)+"\t");

			// iface
			IfaceEntry ie = (IfaceEntry) iface_table.get(n);
			if(ie==null) {
				/* not reachable... */
				System.out.print("-1");
			}
			else {
				System.out.print(ie.iface);
			}
			System.out.println();
		}
	}

	int get_next_iface(UnicastFlow flow)
	{
		Node dest = flow.get_to();
                IfaceEntry ie = (IfaceEntry) iface_table.get(dest);

		if(ie==null) {
			/* not reachable... */
			return -1;
		}
		else {
			if(ie.iface==-1) {
				s.error(this, "iface=-1!");
			}
			return ie.iface;
		}
	}


	public void route(Flow flow) {
		UnicastFlow uflow = (UnicastFlow) flow; // up-cast

		if(uflow.get_to() == node) {
			/* it's for us. stop. */
		}
		else {
			int iface = get_next_iface(uflow);

			if(iface>=0) {
				LinkManager lm = node.get_iface(iface);
				if(lm!=null) ((RoutedFlow)flow).forward(lm);
				else {
					s.error(this,"route: lm=null");
				}
			}
			else {
				if(DEBUG) s.debug(this,"Destination Unreachable: blocking flow.");
				flow.block();
			}
		}
	}

	public DVRouterImpl(Simulator s, Node n, int r) {
		super(s, n,r);
		
		if(DEBUG) s.debug(this,"instantiated.");
		
		ifaces       = n.get_max_ifaces();
		ntables      = new DVTableImpl[ifaces];
		nodes_count  = new int[ifaces];
		cost_table   = new DVTableImpl();
		iface_table  = new Hashtable();

		// insert self into table
		cost_table.put(node,0);
		iface_table.put(node,new IfaceEntry());

		// insert neighbours into table
		for(int i=0; i<ifaces; i++) {
			init_neighbour(i);
		}
		
		run();
	}
}
