package flowsim.slatr;

import flowsim.core.*;
import flowsim.util.Distribution;
import flowsim.util.ExponentialDistribution;

import java.util.Random;

final class FTPTransfer
{
	final Simulator s;
	final FCBEFlow f;
	final DataTransfer dt;

	public int get_remaining_bits() {
		return dt.get_remaining_bits();
	}

	public void start() {
		dt.start();
	}

	public void attach_monitor(DataTransferMonitor dtm) {
		dt.attach_monitor(dtm);
	}

	public FTPTransfer(Simulator s, Node from, int from_port,
			Node to, int to_port, int size, int flow_id, int rport)
	{
		this.s     = s;
		this.f     = new FCBEFlowImpl(s,from,to,flow_id,
				FlowImpl.DS_BEST_EFFORT,rport);
		this.dt    = new DataTransferImpl(s,from_port,to_port,size, f);
	}
}

public class FTPUser extends UserImpl implements DataTransferMonitor
{
	protected final Random r;
	protected final Distribution arrival;
	protected final Distribution size;
	protected final NodeChooser to_chooser;
	protected final int transfers_per_second;
	protected final int average_size;

	int transfers_started;
	int transfers_done;
	long total_duration;

	int flow_id, rport;

	class StartTransferEvent implements Event {
		public void trigger() {
			start_transfer();
			s.schedule((long) arrival.rand(),new StartTransferEvent());
		}
	}

	void start_transfer() {
		Node to;
		do {
			to = to_chooser.get();
		} while(get_node() == to);

		int siz = (int) size.rand();

		/* ftp transfer is made on the opposite direction: you
		 * normally _get_ * files.
		 */
		FTPTransfer ftp = new FTPTransfer(s,to,-1,get_node(),get_port(),
					siz,flow_id,rport);

		ftp.attach_monitor(this);
		ftp.start();
	}

	public void datatransfer_started(DataTransfer dr) {
		max_utility++;
		transfers_started++;
	}
	public void datatransfer_changed_bw(DataTransfer dr) {}
	public void datatransfer_transmitted(DataTransfer dt) {}
	public void datatransfer_arrived(DataTransfer dt) {
		long now = s.now();
		total_duration += now-dt.get_start_time();
		utility++;
		transfers_done++;
	}
	public void datatransfer_blocked(DataTransfer dr) {}

	public void print_stats() {
		System.out.println(this+": budget="+budget+", budget_incr="+budget_incr+
		", max_utility="+get_max_utility()+", utility="+get_utility()+
		", avg. duration="+(get_average_duration()/1000)+"ms");
	}

	public int get_average_duration() {
		return (int)(total_duration/transfers_done);
	}

	/**
	 * @param s Simulator object.
	 * @param node node where the user resides.
	 * @param port port where the user resides.
	 * @param to_chooser ftp transfer destination chooser.
	 * @param budget initial budget.
	 * @param budget_incr budget increase pro second.
	 * @param flow_id flow_id of the generated flows.
	 * @param rport router port by which the ftp transfers should be routed.
	 * @param transfers_per_second number of transfers per second.
	 * @param average_size average size of the transfers.
	 * @param r Random object.
	 */
	public FTPUser(Simulator s, Node node, int port, NodeChooser to_chooser,
		long budget, long budget_incr, int flow_id, int rport,
		int transfers_per_second, int average_size, Random r)
	{
		super(s,node,port,budget,budget_incr);
		this.to_chooser = to_chooser;
		this.flow_id = flow_id;
		this.rport = rport;
		this.transfers_per_second = transfers_per_second;
		this.average_size = average_size;
		this.arrival = new ExponentialDistribution((double)transfers_per_second/1000000,r);
		this.size = new ExponentialDistribution(1.0/average_size,r);
		this.r = r;
		utility = max_utility = 0;
		transfers_started = 0;
		transfers_done = 0;
		total_duration = 0;

		s.schedule((long) arrival.rand(),new StartTransferEvent());
	}
}
