package flowsim.slatr;

import flowsim.core.*;

import com.sun.java.util.collections.Collection;
import com.sun.java.util.collections.HashMap;
import com.sun.java.util.collections.Iterator;
import com.sun.java.util.collections.LinkedList;
import com.sun.java.util.collections.Map;

import java.io.*;

abstract public class SLATRouter
	implements Router, SLABuyer, SLASeller, FlowMonitor, NodeMonitor
{
	static final boolean DEBUG_TRADING = false;
	static final boolean DEBUG_ROUTING = false;
	static final boolean CHECK         = false;
	static final boolean LOG_BUDGET    = true;

	static final long FIXED_COSTS_TIME      = 1000000; // every second
	static final long RELEASE_DS_MARGIN     =  100000; // 100 ms
	static protected final long GIVENSLA_EXPIRATION_MARGIN =  10000; // 10 ms

	static final boolean PERIODIC_PAYMENTS = false;
	static final long PERIODIC_PAYMENTS_PERIOD = 50000;


	protected Simulator s;
	protected Node node;
	protected int port;

	protected SLAMap sla_map;
	protected DSNumbers ds_numbers;

	protected int links_rbw[];
	protected int peer_iface[];
	protected Service peer_service[];
	protected Map peer_number;
	protected int next_peer_nr=0;
	protected SLABuyerProfile buyer_profile[];

	protected Collection routed_flows;

	protected final int peers;
	protected final int ifaces;

	long budget;
	long fixed_costs;

	PrintStream log_budget_ps;

	/********************************/
	/************ EVENTS ************/
	/********************************/

	final class RemoveTakenSlaEvent implements Event {
		final SLA sla;
		public void trigger() {
			//debug("RemoveTakenSlaEvent");
			remove_taken_sla(sla,null);
		}
		RemoveTakenSlaEvent(final SLA sla) { this.sla = sla; }
	}

	final class RemoveTakenBidEvent implements Event {
		final SLA bid;
		public void trigger() {
			//debug("RemoveTakenBidEvent");
			remove_taken_bid(bid,null);
		}
		RemoveTakenBidEvent(final SLA bid) { this.bid = bid; }
	}

	final class RemoveGivenSlaEvent implements Event {
		final SLA sla;
		public void trigger() {
			//debug("RemoveGivenSlaEvent");
			remove_given_sla(sla,null);
		}
		RemoveGivenSlaEvent(final SLA sla) { this.sla = sla; }
	}

	final class RemoveGivenBidEvent implements Event {
		final SLA bid;
		public void trigger() {
			//debug("RemoveGivenBidEvent");
			remove_given_bid(bid,null);
		}
		RemoveGivenBidEvent(final SLA bid) { this.bid = bid; }
	}

	void javac_bug_schedule(long time, Event e) {
		s.schedule(time,e);
	}
	class FixedCostsEvent implements Event {
		public void trigger() {
			s.error(this,"TRIGGERING FIXED COSTS (for the moment should be disabled");
			remove_from_budget(fixed_costs);
			//s.schedule(FIXED_COSTS_TIME, this);
			javac_bug_schedule(FIXED_COSTS_TIME,this);
		}
	}

	class ReleaseDSEvent implements Event {
		int peer, ds;
		Node dest;
		public void trigger() {
			ds_numbers.release_ds(peer, dest, ds);
		}
		ReleaseDSEvent(int peer, Node dest, int ds) {
			this.peer=peer;
			this.dest=dest;
			this.ds=ds;
		}
	}

	/*******************************************/
	/************ SERVICE INTERFACE ************/
	/*******************************************/

	final public Node get_node() {
		return node;
	}

	final public int get_port() {
		return port;
	}

	/*****************************************/
	/************ INTEGRITY CHECK ************/
	/*****************************************/

	final void check() {
		check_links_rbw();
		check_taken_slas_rbw();
	}
	
	final void check_links_rbw() {
		int[] is_ubw;
		is_ubw = new int[ifaces];

		Iterator it = sla_map.taken_slas().iterator();
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			if(sla.iface>=0) {
				is_ubw[sla.iface]+=sla.bw;
			}
			else {
				s.debug(this,"check_links_rbw: sla.iface="+sla.iface);
			}
		}

		for(int i=0; i<ifaces; i++) {
			int is_bw;
			LinkManager lm = node.get_iface(i);
			if(lm==null) {
				is_bw=0;
			}
			else {
				is_bw = lm.get_reservable_bw();
			}

			if(is_bw-is_ubw[i] != links_rbw[i]) {
				s.error(this,"CHECK FAILURE links_rbw: "+
						"is_bw="+is_bw+
						" is_ubw["+i+"]="+is_ubw[i]+
						" links_rbw["+i+"]="+links_rbw[i]);
			}
			if(lm!=null && lm.get_rbw_notbe() < links_rbw[i]) {
				s.error(this,"CHECK FAILURE links_rbw: "+
						" lm.get_rbw_notbe()="+lm.get_rbw_notbe()+
						" links_rbw["+i+"]="+links_rbw[i]);
			}
		}
	}

	final void check_taken_slas_rbw() {
		Iterator it = sla_map.taken_slas().iterator();
		while(it.hasNext()) {
			SLA taken = (SLA) it.next();

			int ubw=0;
			Iterator itb = sla_map.given_slas_to_dest(null);
			while(itb.hasNext()) {
				SLA given = (SLA) itb.next();
				if(sla_map.get_taken_from_given(given)==taken) {
					ubw += given.bw;
				}
			}
			if(taken.bw-ubw!=taken.rbw) {
				s.error(this,"CHECK FAILURE taken_slas_rbw: "+
						"taken.bw="+taken.bw+
						" ubw="+ubw+
						" taken.rbw="+taken.rbw);
				s.error(this," ... taken="+taken);

				itb = sla_map.given_slas_to_dest(null);
				while(itb.hasNext()) {
					SLA given = (SLA) itb.next();
					if(sla_map.get_taken_from_given(given)==taken) {
						s.error(this,"   -> "+given);
					}
				}
			}
		}
	}

	/********************************/
	/************ BUDGET ************/
	/********************************/


	public final void remove_from_budget(long cost) {
		budget -= cost;
		if(DEBUG_TRADING) s.debug(this,"budget = "+budget);
		if(LOG_BUDGET && log_budget_ps!=null) {
			log_budget_ps.println(s.now()+"\t"+budget);
		}
	}
	
	public final void add_to_budget(long gain) {
		budget += gain;
		if(DEBUG_TRADING) s.debug(this,"budget = "+budget);
		if(LOG_BUDGET && log_budget_ps!=null) {
			log_budget_ps.println(s.now()+"\t"+budget);
		}
	}

	/*********************************************/
	/************ PEERS/SLAS DATABASE ************/
	/*********************************************/

	
	long given_expiration(SLA given) {
		int link_delay=0;
		if(given.iface>0) {
			LinkManager lm = node.get_iface(given.iface);
			link_delay+=lm.get_link().get_max_delay();
		}

		return given.expiration+link_delay+GIVENSLA_EXPIRATION_MARGIN;
	}

	void expire_given(SLA given, boolean isbid) {
		final Event e;
		if(isbid) {
			e=new RemoveGivenBidEvent(given);
		}
		else {
			e=new RemoveGivenSlaEvent(given);
		}

		if(!isbid && given.expiration_event!=null) {
			// was a bid. moved to sla -> reschedule for remove_sla.
			given.expiration_event.set_event(e);
			return;
		}

		long exp = given_expiration(given);
		if(exp<s.now()) {
			s.error(this,"'given' already expired!!");
			return;
		}
		
		if(DEBUG_TRADING) s.debug(this,"given "+given+" will expire at "+exp);
		given.expiration_event = s.schedule_absolute(exp,e);
	}

	protected void expire_taken(SLA taken, boolean isbid) {
		if(taken.expiration_event!=null) {
			taken.expiration_event.remove();
		}

		final Event e;
		if(isbid) {
			
			e=new RemoveTakenBidEvent(taken);
		}
		else {
			e=new RemoveTakenSlaEvent(taken);
		}
		
		if(!isbid && taken.expiration_event!=null) {
			// was a bid. moved to sla -> reschedule for remove_sla.
			taken.expiration_event.set_event(e);
			return;
		}

		long exp = taken.expiration;
		if(exp<s.now()) {
			s.error(this,"'taken' already expired!!");
			return;
		}

		taken.expiration_event = s.schedule_absolute(taken.expiration,e);
	}

	protected final int get_peer(Service service) {
		Integer i = (Integer) peer_number.get(service);
		if(i==null) return -1;
		return i.intValue();
	}
	
	protected final int get_peer(Node node, int port) {
		Service service = node.get_service(port);
		if(service==null) return -1;
		return get_peer(service);
	}

	protected void remove_taken_sla(SLA sla, Iterator it) {
		if(sla.rbw!=sla.bw) {
			s.error(this,"removing non-empty taken_sla "+sla);
			Iterator itb = sla_map.given_slas_to_dest(null);
			while(itb.hasNext()) {
				SLA given = (SLA) itb.next();
				if(sla_map.get_taken_from_given(given)==sla) {
					s.error(this,"   -> "+given);
				}
			}
		}
	
		sla_map.remove_taken_sla(sla, it);
		
		int iface = sla.iface;
		release_link_for_sla(sla);
	}
	
	protected void remove_taken_bid(SLA bid, Iterator it) {
		if(DEBUG_TRADING) s.debug(this,"remove_taken_bid: "+bid);
		sla_map.remove_taken_bid(bid, it);
	}

	protected void remove_given_sla(SLA given_sla, Iterator it) {
		// re-add to the taken-sla the bw to rbw.
		SLA taken = (SLA) sla_map.get_taken_from_given(given_sla);

		if(DEBUG_TRADING) s.debug(this,"remove_given_sla "+given_sla+" (taken_sla="+taken+")");

		if(taken!=null) {
			taken.rbw += given_sla.bw;
			if(taken.rbw>taken.bw) {
				s.error(this,"remove_given_sla: rbw="+taken.rbw+" > bw="+taken.bw);
				taken.rbw = taken.bw;
			}
			taken_bw_changed(taken);
		}
		else {
			s.error(this, "remove_given_sla: given="+given_sla+" -> taken=null");
		}

		// delete it
		if(sla_map.remove_given_sla(given_sla, it)) {
			// schedule release of ds-byte
			Event e=new ReleaseDSEvent(given_sla.peer, given_sla.dest, given_sla.ds);
			s.schedule(RELEASE_DS_MARGIN,e);
		}
	}

	protected void remove_given_bid(SLA given_bid, Iterator it) {
		if(DEBUG_TRADING) s.debug(this,"remove_given_bid: "+given_bid);

		// delete it
		if(sla_map.remove_given_bid(given_bid, it)) {
			// schedule release of ds-byte
			Event e=new ReleaseDSEvent(given_bid.peer, given_bid.dest, given_bid.ds);
			s.schedule(RELEASE_DS_MARGIN,e);
		}
	}

	/**********************************/
	/************ MESSAGES ************/
	/**********************************/

	protected void send_ask(SLASeller to, Node dest, int bw, int max_delay, long exp) {
		SLATP_Ask.send(s,this,to,dest,bw,max_delay,exp);
	}

	protected void send_bid(SLABuyer to, SLA sla, SLA bid) {
		//s.debug(this,"Preparing bid "+bid+" based on "+sla+" for "+to);
		if(DEBUG_TRADING) s.debug(this,"Preparing bid based on "+sla+" for "+to);

		// make a copy
		SLA bid_clone;
		try {
			bid_clone = (SLA) bid.clone();
		}
		catch(CloneNotSupportedException e) {
			s.error(this,"make_bid: clone not supported?");
			return;
		}

		// set peer and ds
		int peer = get_peer(to);
		if(peer<0) {
			s.error(this,"make_bid: peer<0");
			return;
		}
		bid_clone.peer = peer;
		bid_clone.iface= peer_iface[peer];
		bid_clone.ds   = ds_numbers.get_ds(peer, bid_clone.dest);

		// put it in the list of given bids
		sla_map.put_given_bid(bid_clone);

		// do the mapping from "given" to "taken"
		sla_map.map_given_to_taken(bid_clone,sla);

		// control expiration
		if(given_expiration(bid_clone)>=sla.expiration) {
			s.error(this,"send_bid: given_expiration(bid_clone)="+given_expiration(bid_clone)+
				", sla.expiration="+sla.expiration);
			return;
		}

		// create expiration event
		expire_given(bid_clone,true);

		// send it
		if(DEBUG_TRADING) s.debug(this,"Sending bid "+bid_clone+" (bw="+bid_clone.bw+")");
		SLATP_Bid.send(s,this,to,bid_clone);
	}

	protected void accept_bid(SLA bid) {
		SLASeller seller = (SLASeller) peer_service[bid.peer];
		if(DEBUG_TRADING) s.debug(this,"Accepting received bid");
		send_accept(seller,bid.dest,bid.ds);
	}

	protected void send_accept(SLASeller to, Node dest, int ds) {
		int peer = get_peer(to);
		if(DEBUG_TRADING) {
			s.debug(this,"Sending accept for bid "+
				"["+peer+": "+dest+", "+ds+"]");
		}
		
		// get bid in question
		SLA bid = sla_map.get_taken_bid(peer,dest,ds);
		if(bid==null) {
			s.error(this,"Attempt to send accept of unknown bid "+
			"["+peer+": "+dest+", "+ds+"]");
			return;
		}
		if(bid.accepted==true) {
			if(DEBUG_TRADING) s.debug(this,"Accept already sent for bid "+bid);
			return;
		}
		bid.accepted=true;
		
		SLATP_Accept.send(s,this,to,dest,ds);
	}

	protected void taken_bw_changed(SLA taken) {}

	protected void send_confirm(SLABuyer to, SLA given) {
		int peer = get_peer(to);

		// move given bid to sla
		sla_map.move_given_bid_to_sla(given);

		// set time to agreement time
		given.time = s.now();

		// create expire event
		expire_given(given,false);

		SLA taken = sla_map.get_taken_from_given(given);
		if(taken==null) {
			s.error(this,"send_confirm: Unknown taken SLA!");
			return;
		}

		// get money for sold service...
		if(!PERIODIC_PAYMENTS) add_to_budget(given.cost);

		// control expiration
		if(given_expiration(given)>=taken.expiration) {
			s.error(this,"send_confirm: given_expiration(to)="+given_expiration(given)+
				", taken.expiration="+taken.expiration);
			return;
		}

		if(DEBUG_TRADING) {
			s.debug(this,"Sending confirm for sla "+given+" (taken="+taken+")");
		}
		
		// subtract bw from rbw of taken
		if(DEBUG_TRADING) s.debug(this,"given.bw="+given.bw+"-> taken.rbw="+taken.rbw);
		taken.rbw -= given.bw;
		if(taken.rbw<0) {
			s.error(this,"make_bid: taken.rbw<0");
			return;
		}

		taken_bw_changed(taken);

		SLATP_Confirm.send(s,this,to,given.dest,given.ds);
	}

	protected void send_reject(SLABuyer to, Node dest, int ds) {
		if(DEBUG_TRADING) {
			int peer = get_peer(to);
			s.debug(this,"Sending reject for sla ["+peer+": "+dest+", "+ds+"]");
		}
		SLATP_Reject.send(s,this,to,dest,ds);
	}

	/*****************************************/
	/************ BUYER INTERFACE ************/
	/*****************************************/

	
	public SLABuyerProfile get_profile() {
		return new SLABuyerProfile(true,false,true);
	}

	abstract protected void bid_received(SLA bid);

	public final void receive_bid(SLATP_Bid msg) {
		if(msg.sla.dest==node) {
			s.error(this,"router "+msg.get_from()+" sent me a bid to myself!");
			return; // It's me!
		}

		if(msg.sla.expiration<s.now()) {
			s.error(this,"bid already expired!");
			return;
		}

		msg.sla.peer  = get_peer(msg.get_from());
		msg.sla.iface = peer_iface[msg.sla.peer];
		msg.sla.accepted=false;

		if(msg.sla.peer<0) {
			s.error(this,"msg.sla.peer<0!");
			return;
		}
		
		sla_map.put_taken_bid(msg.sla);

		if(DEBUG_TRADING) s.debug(this,"Received bid "+msg.sla+
			" from "+msg.get_from()+
			" (iface="+msg.sla.iface+")");

		expire_taken(msg.sla,true);

		bid_received(msg.sla);
	}

	protected abstract void confirm_received(SLA sla);

	public void receive_confirm(SLATP_Confirm msg) {
		int peer = get_peer(msg.get_from());

		if(DEBUG_TRADING) s.debug(this,"Confirm received for sla ["+peer+": "+
			msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");

		SLA bid = sla_map.get_taken_bid(peer,msg.get_sla_dest(),msg.get_sla_ds());

		// Check if an accept message was sent or if it is junk...
		if(!bid.accepted) {
			s.error(this,"Junk SLAT_Confirm: "+
				"["+peer+": "+msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");
			return;
		}

		// set time to agreement time
		bid.time = s.now();

		// detract spent money
		if(!PERIODIC_PAYMENTS) {
			if(node.get_name().equals("5")) {
				s.debug(this,"remove from budget: "+bid.cost);
			}
			remove_from_budget(bid.cost);
		}

		if(!reserve_link_for_sla(bid)) {
			// throw away taken sla, because insufficent
			// bw on iface (should be avoided by intelligent traders)
			if(DEBUG_TRADING) {
				s.debug(this, "WARNING: confirm received, but bw "+
						"insufficient on link (bid="+bid+")");
			}
			sla_map.remove_taken_bid(bid);
			return;
		}

		sla_map.move_taken_bid_to_sla(bid);

		expire_taken(bid,false);

		confirm_received(bid);

		if(CHECK) check();
	}

	protected abstract void reject_received(SLA sla);

	public void receive_reject(SLATP_Reject msg) {
		int peer = get_peer(msg.get_from());

		if(DEBUG_TRADING) s.debug(this,"Reject received for sla ["+peer+": "+
			msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");

		//s.debug(this,"Reject received for sla ["+peer+": "+
		//	msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");

		SLA bid = sla_map.get_taken_bid(peer,msg.get_sla_dest(),msg.get_sla_ds());

		// Check if an accept message was sent or if it is junk...
		if(!bid.accepted) {
			s.error(this,"Junk SLAT_Reject message received!");
			return;
		}

	
		if(bid.expiration_event!=null) {
			bid.expiration_event.remove();
		}
		sla_map.remove_taken_bid(bid);

		reject_received(bid);
	}

	/******************************************/
	/************ SELLER INTERFACE ************/
	/******************************************/

	public void add_buyer(SLABuyer buyer, SLABuyerProfile profile) {
		if(DEBUG_TRADING) s.debug(this,"adding buyer "+buyer+" (peer="+next_peer_nr+")");
	
		if(next_peer_nr+1 >= peers) return;
		peer_service[next_peer_nr] = buyer;
		peer_number.put(buyer, new Integer(next_peer_nr));
		buyer_profile[next_peer_nr] = profile;

		final Node bnode = buyer.get_node();
		if(bnode!=node) {
			peer_iface[next_peer_nr] = node.get_iface(bnode);
		}
		else {
			peer_iface[next_peer_nr] = -1;
		}

		next_peer_nr++;
	}

	public void receive_ask(SLATP_Ask ask) {}

	public abstract boolean send_confirm_for_accept(SLA sla);

	protected void confirm_sent(Service to, SLA sla) {}

	protected void reject_sent(Service to, SLA sla) {}

	public void receive_accept(SLATP_Accept msg) {
		// move from given_bids to given_slas and send confirmation
		
		int peer = get_peer(msg.get_from());

		if(DEBUG_TRADING) {
			s.debug(this,"Accept received for bid ["+
				peer+": "+msg.sla_dest+", "+msg.sla_ds+"]");
		}
		SLA sla = sla_map.get_given_bid(peer, msg.sla_dest, msg.sla_ds);
		if(sla!=null) {
			// first come, first serve...
			if(send_confirm_for_accept(sla)) {
				send_confirm((SLABuyer) msg.get_from(),sla);
				confirm_sent(msg.get_from(),sla);
			}
			else {
				sla_map.remove_given_bid(sla,null);
				send_reject((SLABuyer) msg.get_from(),msg.sla_dest,msg.sla_ds);
				reject_sent(msg.get_from(),sla);
			}
		}
		else {
			s.error(this,"receive_accept: bid "+
				"["+peer+": "+msg.sla_dest+", "+msg.sla_ds+"]"+
				" not in given_bids??");
		}
	}

	/*****************************************/
	/************ LINK MANAGEMENT ************/
	/*****************************************/

	protected final boolean reserve_link_for_sla(SLA sla) {
		if(DEBUG_TRADING) s.debug(this,"reserve "+sla);
		// adjust rbw on link
		int iface = sla.iface;
		if(iface<0) {
			s.error(this,"reserve_link_for_sla: iface<0");
			return false;
		}

		if(links_rbw[iface]<sla.bw) {
			if(DEBUG_TRADING) s.debug(this, "Warning: reserve_link_for_sla failed "+
				"(sla="+sla+", "+links_rbw[iface]+")");
			return false;
		}
		//s.debug(this,"before: links_rbw["+iface+"]="+links_rbw[iface]);
		links_rbw[iface] -= sla.bw;
		//s.debug(this,"after : links_rbw["+iface+"]="+links_rbw[iface]);

		return true;
	}

	protected final void release_link_for_sla(SLA sla) {
		if(DEBUG_TRADING) s.debug(this,"release "+sla);
		int iface = sla.iface;
		if(iface>=0) {
			links_rbw[iface] += sla.bw;
			if(links_rbw[iface]>node.get_iface(iface).get_reservable_bw()) {
				s.error(this,"remove_taken_sla: rbw>bw");
			}
		}
		else {
			s.error(this,"taken_sla from local SLASeller?");
		}
	}

	/**************************************/
	/************ FLOW MONITOR ************/
	/**************************************/

	void readd_given_sla_rbw(Flow _flow) {
		UnicastFlow flow = (UnicastFlow) _flow;
		int peer = get_source_peer_of_flow(flow.get_child());
		if(peer<0) {
			s.error(this,"readd_given_sla: peer<0!");
			return;
		}

		SLA given = sla_map.get_given_sla(peer, flow.get_to(), flow.get_ds());
		if(given==null) {
			return;
		}

		//s.debug(this, "Before: "+given);
		given.rbw+=flow.get_bw();
		//s.debug(this, "After : "+given);
		if(given.rbw>given.bw) {
			s.error(this,"readd_given_sla_rbw: "+
				"given.rbw="+given.rbw+
				", given.bw="+given.bw);
		}
	}

	public void flow_started(Flow _flow) {}

	public void flow_stopped(Flow _flow) {
		UnicastFlow flow = (UnicastFlow) _flow;
		if(DEBUG_ROUTING) s.debug(this,"flow "+flow+" stopped");
		routed_flows.remove(flow);
		readd_given_sla_rbw(flow);
	}

	public void flow_blocked(Flow _flow) {
		UnicastFlow flow = (UnicastFlow) _flow;
		if(DEBUG_ROUTING) s.debug(this,"flow "+flow+" blocked");
		routed_flows.remove(flow);
		readd_given_sla_rbw(flow);
	}

	public void flow_changed_bw(Flow f, int was_bw) {
		s.debug(this,"changed bw");
		UnicastFlow flow = (UnicastFlow) f;
		int peer = get_source_peer_of_flow(flow.get_child());
		if(peer<0) {
			s.error(this,"remove_given_sla: peer<0!");
			return;
		}

		SLA given = sla_map.get_given_sla(peer, flow.get_to(), flow.get_ds());
		if(given==null) {
			return;
		}

		// remove old bandwidth
		given.rbw+=was_bw;
		if(given.rbw>given.bw) {
			s.error(this,"flow_changed_bw: "+
				"given.rbw="+given.rbw+
				", given.bw="+given.bw);
			return;
		}

		// add new bandwidth
		if(given.rbw>f.get_bw()) {
			given.rbw-=f.get_bw();
		}
		else {
			if(DEBUG_ROUTING) s.debug(this,"flow changed bandwidth and "+
				"exceeded capacity -> blocking it.");
			f.block();
			return;
		}
	}

	public void flow_changed_desired_bw(Flow f, int was_bw) {}
	public void flow_changed_available_bw(Flow f, int was_bw) {}


	/******************************************/
	/************ ROUTER INTERFACE ************/
	/******************************************/

	final int get_source_peer_of_flow(UnicastFlow flow) {
		Service service = flow.get_from_service();
		return get_peer(service);
	}

	boolean make_child(UnicastFlow flow) {
		if(DEBUG_ROUTING) s.debug(this,"make_child (flow="+flow+", "+"parent="+flow.get_parent()+")");
		if(flow.get_parent()!=null &&
			routed_flows.contains(flow.get_parent())) return false;

		Service service = flow.get_from_service();
		if(service==null) {
			if(flow.get_parent()==null) {
				s.error(this,"local peer didn't set from_service of flow!");
				flow.block();
				return true;
			}
			service = flow.get_parent().get_from().get_service(get_port());
		}

		Flow child = flow.add_child_same_node();

		if(DEBUG_ROUTING) s.debug(this,"Adding child on same node -> "+child);

		routed_flows.add(flow);

		child.set_from_service(service);
		child.start();

		return true;
	}

	public void route(Flow _flow) {
		UnicastFlow flow = (UnicastFlow) _flow; // up-cast

		if(flow.get_to() == node) return; // it's for us...

		if(DEBUG_ROUTING) s.debug(this,"route "+_flow.hashCode()+
				" (from="+flow.get_from()+", to="+flow.get_to()+")");

		if(make_child(flow)) return;

		if(DEBUG_ROUTING) s.debug(this,"route "+_flow.hashCode()+
				" (from="+flow.get_from()+", to="+flow.get_to()+")");

		int peer = get_source_peer_of_flow(flow);
		if(peer<0) {
			s.error(this,"route: peer<0!");
			flow.block();
			routed_flows.remove(flow.get_parent());
			return;
		}

		SLA given = sla_map.get_given_sla(peer, flow.get_to(), flow.get_ds());
		if(given==null) {
			s.error(this,"route: given=null");	
			flow.block();
			routed_flows.remove(flow.get_parent());
			return;
		}

		SLA taken = sla_map.get_taken_from_given(given);
		if(taken==null) {
			s.error(this,"route: taken=null");	
			flow.block();
			routed_flows.remove(flow.get_parent());
			return;
		}

		if(taken.iface<0||taken.iface>=ifaces) {
			s.error(this,"route: taken.iface="+taken.iface);
			flow.block();
			routed_flows.remove(flow.get_parent());
			return;
		}

		if(given.rbw<flow.get_bw()) {
			s.error(this,"Flow "+flow+" out of profile (given.rbw="+given.rbw+")");
			if(DEBUG_ROUTING) s.debug(this,"flow="+flow+", given="+given+")");
			flow.block();
			routed_flows.remove(flow.get_parent());

			debug_table();
			SLATRouter r = (SLATRouter) peer_service[given.peer];
			r.debug_table();

			return;
		}

		flow.set_ds(taken.ds);
		//s.debug(this, "Before: "+given);
		given.rbw-=flow.get_bw();
		//s.debug(this, "After : "+given);

		// parent is the real incoming flow...
		flow.get_parent().attach_monitor(this);

		if(DEBUG_ROUTING) s.debug(this,"Forwarding "+flow+" to "+
			taken.iface+" (given="+given+")");

		((RoutedFlow)flow).forward(node.get_iface(taken.iface));

		if(DEBUG_ROUTING) s.debug(this,"Routing done");
	}

	public void iface_changed(Node n, int i) {
		// assumption: neighbour was _added_

		// set links_rbw[i]
		LinkManager lm = node.get_iface(i);
		if(lm==null) {
			links_rbw[i]=0;
			return;
		}
		links_rbw[i] = lm.get_reservable_bw();

		// setup peer on same port
		SLABuyer buyer = (SLABuyer) lm.get_peer().get_service(get_port());

		// FIXME: this could be made into SLATP protocol...
		// it does fail if links are created before the routers!!!!!
		if(buyer!=null) {
			add_buyer(buyer,buyer.get_profile());
		}
	}

	public void start() {
		for(int i=0; i<ifaces; i++) {
			iface_changed(node,i);
		}
	}


	/***************************************/
	/************ DEBUG HELPERS ************/
	/***************************************/

	public String toString() {
		return "SLATRouter("+get_node()+")";
	}

	public void debug_table() {
		debug_table(System.out);
	}

	public void debug_table(PrintStream ps) {
		Iterator e;

		ps.println(this+" budget="+budget);
		ps.println("===============================================================================");
		
		ps.println("\trouted flows:");
		e = routed_flows.iterator();
		while(e.hasNext()) {
			UnicastFlow flow = (UnicastFlow) e.next();
			ps.print("\t\t"+flow);
			if(flow.get_parent()!=null) {
				ps.print(" <- "+flow.get_parent());
				if(flow.get_parent().get_parent()!=null) {
					ps.print(" <- "+
						flow.get_parent().get_parent());
				}
			}
			ps.println();
		}
		
		ps.println("\ttaken bids:");
		e = sla_map.taken_bids().iterator();
		while(e.hasNext()) {
			SLA sla = (SLA) e.next();
			ps.println("\t\t"+sla.toDebugString());
		}

		ps.println("\ttaken slas:");
		e = sla_map.taken_slas().iterator();
		while(e.hasNext()) {
			SLA sla = (SLA) e.next();
			ps.println("\t\t"+sla.toDebugString());
		}

		for(int i=0; i<peers; i++) {
			Service service = peer_service[i];
			if(service==null) continue;

			ps.print(i+": node "+ service.get_node()+
				" port "+service.get_port());
			if(service instanceof SLATRouter) {
				int iface = node.get_iface(service.get_node());
				int total_bw = node.get_iface(iface).get_reservable_bw();
				int rbw_percent = (int)(((long)links_rbw[iface])*1000/total_bw);
				ps.println(": SLATRouter: rbw "+links_rbw[iface]+
						"("+((float)rbw_percent)/10+"%)");
			}
			else if(service instanceof User) {
				User u = (User) service;
				ps.println(": User: max_utility "+u.get_max_utility()+
					" utility "+u.get_utility());
			}
			else {
				ps.println();
			}

			ps.println("\tgiven bids:");
			e = sla_map.given_bids(i).iterator();
			while(e.hasNext()) {
				SLA sla = (SLA) e.next();
				ps.println("\t\t"+sla.toDebugString());
			}

			ps.println("\tgiven slas:");
			e = sla_map.given_slas(i).iterator();
			while(e.hasNext()) {
				SLA sla = (SLA) e.next();
				ps.println("\t\t"+sla.toDebugString());
			}
			ps.println();
		}
	}

	final class PeriodicPaymentsEvent implements Event {
		public void trigger() { periodic_payments(this); }
	}

	void periodic_payments(Event e) {
		if(e==null) e=new PeriodicPaymentsEvent();

		Iterator it;
		float t=(float)PERIODIC_PAYMENTS_PERIOD/1000000;

		// income
		long income=0;
		for(int i=0; i<peers; i++) {
			Service service = peer_service[i];
			if(service==null) continue;

			it = sla_map.given_slas(i).iterator();
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				income += (int)(sla.cost_time()*t);
			}
		}
		add_to_budget(income);

		// expenses
		long expenses=0;
		it = sla_map.taken_slas().iterator();
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			expenses += (int)(sla.cost_time()*t);
		}
		remove_from_budget(expenses);

		s.schedule(PERIODIC_PAYMENTS_PERIOD, e);
	}

	/*************************************/
	/************ CONSTRUCTOR ************/
	/*************************************/

	public SLATRouter(Simulator s, Node node, int port,
		int local_peers, long budget, long fixed_costs) {

		this.s      = s;
		this.node   = node;
		this.port   = port;
		this.ifaces = node.get_max_ifaces();
		this.peers  = ifaces+local_peers;
		this.sla_map = new SLAMap(s,peers,get_node().toString());
		this.ds_numbers = new DSNumbers(s,node.toString());
		
		// router init on node
		node.set_service(this,port);
		node.attach_monitor(this);

		// init links_rbw
		this.links_rbw = new int[ifaces];
		for(int i=0; i<ifaces; i++) {
			LinkManager lm = node.get_iface(i);
			if(lm==null) {
				links_rbw[i]=0;
				continue;
			}
			links_rbw[i] = lm.get_reservable_bw();
		}

		// peers
		this.peer_service = new Service[peers];
		this.peer_number = new HashMap();
		this.peer_iface  = new int[peers];
		this.buyer_profile = new SLABuyerProfile[peers];

		// budget & fixed costs
		this.budget = budget;
		/*
		this.fixed_costs = fixed_costs;
		s.schedule(FIXED_COSTS_TIME, new FixedCostsEvent());
		*/


		// routing
		routed_flows = new LinkedList();

		if(PERIODIC_PAYMENTS) periodic_payments(null);

		// budget log
		log_budget_ps = null;
		try {
			if(LOG_BUDGET) {
				FileOutputStream fos = new FileOutputStream("logs/budget"+
						"-"+node.get_name()+".log");
				log_budget_ps = new PrintStream(fos);
			}
		} catch(IOException e) {}
	}
}
