package flowsim.slatr;

import flowsim.core.*;
import flowsim.util.FloatBuffer;
import flowsim.util.Log2;

import java.util.Random;

import com.sun.java.util.collections.Collection;
import com.sun.java.util.collections.Iterator;

public abstract class SLATRouterImpl extends SLATRouter
{
	static final boolean DEBUG_IMPL = false;

	static final long BASIS_CYCLE_TIME    = 100000;	// 100ms

	static final long NEIGHBOUR_EXPIRE    = 1000*BASIS_CYCLE_TIME;
	static final float NEIGHBOUR_COST     = (float) 0.01; // 1 cent per bit
	static final long NEIGHBOUR_FIX_COST  = 10000;

	static final long TAKENSLA_EXPIRATION_MARGIN = BASIS_CYCLE_TIME;

	static final long MIN_BID_EXPIRATION  = 20*BASIS_CYCLE_TIME;
	static final long MAX_BID_EXPIRATION  = 1000*BASIS_CYCLE_TIME;

	static final long NEWBID_EXPIRATION_MARGIN = MIN_BID_EXPIRATION;
	static final float NEWBID_PRICE_MARGIN = 0.8f;

	static final long PRICE_FIXED         = 100;
	static final float PRICE_MAX_FACTOR  = 2.0f;
	static final float PRICE_MIN_FACTOR  = 1.5f;

	static final int DELAY_CLASS_THRESHOLD = 200000;

	// the link delay is also added
	static final int  ADDED_DELAY         = 10000;
	// start with bids with that bw
	static final int  BID_MIN_BW          = 64000;
	// start with bids with that bw
	static final int  BID_MAX_BW          = 512000;

	// Accepts for bids of 80% of current price will be accepted.
	static final double  PRICE_TOLERANCE = 0.8;

	static final long PRICE_INFINITY = Long.MAX_VALUE;

	static final boolean SUBSTRACT_BW_BEFORE_PRICE = false;
	static final boolean PRICE_LINK_RBW_DEPENDENT = true;


	Random random;

	final class RemoveTakenSlaEvent implements Event {
		final SLA sla;
		public void trigger() {
			//debug("RemoveTakenSlaEvent");
			remove_taken_sla(sla,null);
		}
		RemoveTakenSlaEvent(final SLA sla) { this.sla = sla; }
	}

	class RunEvent implements Event {
		public void trigger() { run(); }
	}

	Iterator search_taken_sla(Node dest, int max_delay) {
		return sla_map.search_taken_slas(dest,
				-1, -1,
				-1, -1,
				-1, -1,
				-1, max_delay,
				-1L, -1L,
				s.now()+TAKENSLA_EXPIRATION_MARGIN, -1);
	}

	Iterator search_taken_sla_same_class(Node dest, int delay) {
		return search_taken_sla(dest,dlclass_max(dlclass(delay)));
	}

	protected long get_bw_taken_sla(Node dest, int max_delay) {
		long total_bw=0;
		{
			Iterator it = search_taken_sla(dest,max_delay);
			while(it.hasNext()) {
				SLA t = (SLA) it.next();
				total_bw     += t.bw;
			}
		}
		return total_bw;
	}

	protected long get_bw_taken_sla_same_class(Node dest, int delay) {
		return get_bw_taken_sla(dest,dlclass(delay));
	}

	protected float get_rbwbw_taken_sla(Node dest, int max_delay) {
		float rbwbw;
		{
			long total_bw=0, total_rbw=0;
			Iterator it = search_taken_sla(dest,max_delay);
			while(it.hasNext()) {
				SLA t = (SLA) it.next();
				total_bw     += t.bw;
				total_rbw    += t.rbw;
			}
			if(total_bw==0) {
				rbwbw = (float) 0;
			}
			else {
				rbwbw = (float) total_rbw/total_bw;
			}
		}
		return rbwbw;
	}

	protected float get_rbwbw_taken_sla_same_class(Node dest, int delay) {
		return get_rbwbw_taken_sla(dest,dlclass_max(dlclass(delay)));
	}

	protected float price_volume(Node dest, long volume, int bw, int delay)
	{
		float rbwbw;
		float cost;
		{
			long total_bw=0, total_rbw=0;
			long total_volume=0, total_cost=0;
			long now = s.now();

			Iterator it = search_taken_sla_same_class(dest,delay);
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				total_volume += sla.volume(now);
				total_cost   += sla.cost;

				if(!PRICE_LINK_RBW_DEPENDENT || sla.iface<0) {
					total_bw  += sla.bw;
					total_rbw += sla.rbw;
				}
				else {
					total_bw  += node.get_iface(sla.iface).get_reservable_bw();
					total_rbw += links_rbw[sla.iface];
				}
			}
			//s.debug(this,"total_bw="+total_bw+", total_rbw="+total_rbw);
			if(total_bw==0) {
				rbwbw = (float) 0;
				cost  =  0;
			}
			else {
				if(SUBSTRACT_BW_BEFORE_PRICE) {
					total_rbw -= bw;
					if(total_rbw<0) total_rbw=0;
				}

				rbwbw = (float) total_rbw/total_bw;
				cost  = (float) total_cost/total_volume;
			}
		}

		if(rbwbw>1 || cost<0) {
			s.error(this,"rbwbw="+rbwbw+", cost="+cost);
			return Float.MAX_VALUE;
		}

		float max_per_volume   = PRICE_MAX_FACTOR*cost;
		float min_per_volume   = PRICE_MIN_FACTOR*cost;

		float price_per_volume = (min_per_volume +
			(1-rbwbw)*(max_per_volume-min_per_volume));

		if(DEBUG_IMPL) s.debug(this,"cost="+cost+" => price="+price_per_volume+ " (rbwbw="+rbwbw+")");

		return price_per_volume;
	}

	public long price(Node dest, long volume, int delay, int bw) {
		return PRICE_FIXED+(long)(price_volume(dest,volume,delay,bw)*volume);
	}

	public void receive_ask(SLATP_Ask msg) {
		if(DEBUG_IMPL) {
			s.debug(this,"Ask received "+
				//"from "+msg.get_from()+" "+
				"("+msg.dest+" "+
				+msg.bw/1000+"kb "+
				+msg.max_delay/1000+"ms "+
				"exp="+msg.expiration+")"
			);
		}

		send_best_bid((SLABuyer) msg.get_from(), msg.dest, msg.bw,
			msg.max_delay, msg.expiration);
	}

	protected void send_best_bid(SLABuyer buyer, Node dest,
		int bw, int max_delay, long exp)
	{
		//s.debug(this,"send_best_bid");
		int added_delay = ADDED_DELAY;
		if(buyer.get_node()!=node) {
			int iface = node.get_iface(buyer.get_node());
			LinkManager lm = node.get_iface(iface);
			if(lm==null) return;
			Link l = lm.get_link();

			added_delay+=l.get_max_delay();
		}


		Iterator it = sla_map.search_taken_slas(dest,
				-1, -1, // bw
				bw, -1, // rbw
				-1, -1, // rbwbw
				-1, max_delay-added_delay, // delay
				-1, -1, // cost
				exp+GIVENSLA_EXPIRATION_MARGIN+added_delay, -1 // exp
				);

		SLA best_bid=null;
		SLA best_sla=null;
		long least_cost=Long.MAX_VALUE;
		final long now = s.now();
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			SLA bid = new SLA(dest,
					-1,bw,
					sla.max_delay+added_delay,
					-1,s.now(),exp);
			bid.cost = price(bid.dest,bid.volume(),bid.bw,bid.max_delay);
			if(bid.cost<least_cost) {
				least_cost = bid.cost;
				best_bid = bid;
				best_sla = sla;
			}
		}

		// if appropriate bid could be created, send it...
		if(best_bid!=null) {
			send_bid(buyer,best_sla,best_bid);
		}
	}

	public boolean send_confirm_for_accept(SLA given) {
		// verify if the resources are there...
		SLA taken = sla_map.get_taken_from_given(given);
		if(taken==null) {
			if(DEBUG_IMPL) s.debug(s,"accept received for bid "+given+
				" but taken-service not found.");
			return false;
		}
		if(taken.rbw<given.bw) {
			if(DEBUG_IMPL) s.debug(s,"accept received for bid "+given+
				" but taken.rbw<given.bw (taken="+taken+")");
			return false;
		}
		
		// verify price
		/*
		if(given.cost<price(given.dest,given.volume(),
			given.bw,given.max_delay)*PRICE_TOLERANCE) return false;
		*/
		
		return true;
	}

	// delay class
	int dlclass(final int dl) {
		if(dl>DELAY_CLASS_THRESHOLD) return 1;
		else return 0;
	}
	
	int dlclass_max(final int dlc) {
		if(dlc>=1) return Integer.MAX_VALUE;
		else return DELAY_CLASS_THRESHOLD;
	}

	final public boolean has_link_space(SLA bid) {
		if(links_rbw[bid.iface]>bid.bw) return true;
		return false;
	}

	abstract protected void bid_received(SLA bid);

	protected void confirm_received(SLA sla) {
		make_bids_on_taken_sla(sla);
	}

	protected void confirm_sent(Service to, SLA sla) {
		// make another offer like this one.
		//send_best_bid((SLABuyer) to,sla.dest,sla.bw,sla.max_delay,sla.expiration);
		make_bids();
	}

	protected void reject_sent(Service to, SLA sla) {
		// make another offer like this one.
		//send_best_bid((SLABuyer) to,sla.dest,sla.bw,sla.max_delay,sla.expiration);
		make_bids();
	}

	public void reject_received(SLA sla) {}

	void make_bids_on_taken_sla(SLA sla) {
		for(int i=0; i<peers; i++) {
			if(i!=sla.peer);
			for(int bw=BID_MIN_BW; bw<=BID_MAX_BW; bw*=2) {
				make_bid_on_taken_sla(sla,i,bw);
			}
		}
	}

	void make_bid_on_taken_sla(SLA sla, int peer, int bw) {
		SLABuyer buyer = (SLABuyer) peer_service[peer];

		if(buyer==null) return;
		if(bw>sla.rbw) return;
		if(bw<BID_MIN_BW) return;
		if(buyer.get_node()==sla.dest) return;

		// profile: send not asked bids?
		SLABuyerProfile profile = buyer_profile[peer];
		if(profile==null) {
			s.error(this,"profile for buyer not set?");
			return;
		}
		if(!profile.send_not_asked_bids) {
			if(DEBUG_IMPL) {
				s.debug(this,"Buyer "+buyer+" doesn't want not asked bids...\n");
			}
			return;
		}

		int link_delay = 0;
		if(peer_iface[peer]>0) {
			LinkManager lm = node.get_iface(peer_iface[peer]);
			link_delay+=lm.get_link().get_max_delay();
		}
		long       now = s.now();
		long       exp = sla.expiration-link_delay-GIVENSLA_EXPIRATION_MARGIN-1;

		if(exp<now+MIN_BID_EXPIRATION) return;

		if(exp>now+MAX_BID_EXPIRATION) {
			exp = now+MAX_BID_EXPIRATION;
		}

		SLA bid = new SLA(sla.dest,-1,bw,
				sla.max_delay+ADDED_DELAY+link_delay,
				-1, now, exp, peer);
		bid.cost  = price(bid.dest,bid.volume(),bid.bw,bid.max_delay);

		send_bid_if_new(buyer,sla,bid);
	}

	boolean is_bid_new(SLABuyer to, SLA bid) {
		// check if there is already such a bid in the given_bids list
		// of that iface...
		int peer = get_peer(to);
		Iterator it = sla_map.search_given_bids(peer, bid.dest,
				//bwclass_min(bwclass(bid.bw)), -1,
				bid.bw, bid.bw+1,
				-1, -1, // rbw
				-1, -1, // rbwbw
				-1, dlclass_max(dlclass(bid.max_delay))+1,
				-1, -1,
				bid.expiration-NEWBID_EXPIRATION_MARGIN, -1);

		while(it.hasNext()) {
			SLA old_bid = (SLA) it.next();
			// is this price much better than what was already
			// sent?
			long now = s.now();
			if(old_bid.cost_volume(now)*NEWBID_PRICE_MARGIN >
				bid.cost_volume(now))
			{
				return true;
			}
			//s.debug(this,"A good bid was already sent...");
			return false;
		}

		return true;
	}

	void send_bid_if_new(SLABuyer to, SLA from, SLA bid) {
		if(is_bid_new(to,bid)) {
			send_bid(to,from,bid);
		}
	}

	protected void make_bids() {
		if(DEBUG_IMPL) s.debug(this,"make_bids()");
		
		Iterator it=sla_map.search_taken_slas(null,
				-1, -1,
				1000, -1, // at least 1k remaining
				-1, -1,
				-1, -1,
				-1L,-1L,
				s.now()+MIN_BID_EXPIRATION+GIVENSLA_EXPIRATION_MARGIN, -1);

		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			make_bids_on_taken_sla(sla);
		}
	}

	protected void buy_slas() {}

	protected void run() {
		buy_slas();
		make_bids();
		s.schedule(BASIS_CYCLE_TIME, new RunEvent());
	}

	public void iface_changed(Node n, int i) {
		super.iface_changed(n,i);
		add_neighbour_bids(i);
	}

	protected void accept_bid(SLA bid) {
		if(bid.peer==-1) {
			// neighbour_bid
			if(DEBUG_IMPL) s.debug(this,"accept_bid for neighbour bid");

			if(!reserve_link_for_sla(bid)) {
				// throw away taken sla, because insufficent
				// bw on iface (should be avoided by intelligent traders)
				if(DEBUG_IMPL) {
					s.debug(this, "WARNING: confirm received, but bw "+
							"insufficient on link (bid="+bid+")");
				}
				sla_map.remove_taken_bid(bid);
				return;
			}

			sla_map.move_taken_bid_to_sla(bid);

			expire_taken(bid,false);

			// readd new bid
			add_neighbour_bid(bid.iface, bid.bw);

			confirm_received(bid);

			// if(CHECK) check();
		}
		else {
			super.accept_bid(bid);
		}
	}

	protected void add_neighbour_bids(int iface) {
		add_neighbour_bid(iface,64000);
		add_neighbour_bid(iface,128000);
		add_neighbour_bid(iface,256000);
		add_neighbour_bid(iface,512000);
		add_neighbour_bid(iface,1024000);
	}

	protected void add_neighbour_bid(int iface, int bw) {
		LinkManager lm = node.get_iface(iface);
		if(lm==null) return;

		Node neighbour = lm.get_peer();
		long exp       = s.now()+NEIGHBOUR_EXPIRE;
		
		SLA sla = new SLA(neighbour, -1, bw,
			0, 0, s.now(), exp, -1);
		sla.cost = (long)(NEIGHBOUR_COST*sla.volume()+NEIGHBOUR_FIX_COST);
		sla.ds = ds_numbers.get_ds(sla.peer, sla.dest);
		sla.iface=iface;

		sla_map.put_taken_bid(sla);

		// schedule expiration
		expire_taken(sla,true);

		bid_received(sla);
	}

	public void start() {
		super.start();

		long run_time = (long)(random.nextFloat()*BASIS_CYCLE_TIME);
		s.schedule(run_time, new RunEvent());
	}

	public String toString() {
		return "SLATRouterGreedy("+get_node()+")";
	}

	public SLATRouterImpl(Simulator s, Node node, int port,
		int local_peers, long budget, long fixed_costs, Random random)
	{
		super(s,node,port,local_peers,budget,fixed_costs);

		this.random = random;

		// add as 'taken_slas' with cost 0 possible services to
		// neighbours
		for(int i=0; i<node.get_max_ifaces(); i++) {
			add_neighbour_bids(i);
		}

		start();
	}
}
