package flowsim.slatr;

import flowsim.core.*;

import java.util.Random;

import com.sun.java.util.collections.Set;
import com.sun.java.util.collections.HashSet;
import com.sun.java.util.collections.Iterator;

import java.io.*;

public class SLATRouterLazy extends SLATRouterImpl
{
	static final boolean DEBUG_LAZY = false;

	// bids with an expiration less than this won't be even considered.
	static final long TAKENBID_MIN_EXPIRATION = BASIS_CYCLE_TIME;

	static final int ASK_WAIT_TIME = 100000;
	static final int MIN_DELAY = 10000; // if delay should be less than that, don't bother asking.

	static final long ASK_MIN_EXPIRATION = 10*BASIS_CYCLE_TIME;
	static final long ASK_MIN_BANDWIDTH  = 64000;

	// buy no more as 100% of already available bw.
	static final int   TAKENSLAS_MAX_INCREASE = 512000;
	static final float TAKENSLAS_MAX_INCREASE_BW = 0.5f;
	static final int   TAKENSLAS_MIN_INCREASE = 64000;

	SLAList asked;

	final class BuyEvent implements Event {
		final Node dest;
		final int bw, max_delay;
		final long exp;
		public void trigger() {
			buy_best(dest,bw,max_delay,exp);
		}
		BuyEvent(Node dest, int bw, int max_delay, long exp) {
			this.dest=dest;
			this.bw=bw;
			this.max_delay=max_delay;
			this.exp=exp;
		}
	}

	final class RemoveFromAskedEvent implements Event {
		final SLA sla;
		public void trigger() {
			asked.remove(sla);
			ds_numbers.release_ds(sla.peer,sla.dest,sla.ds);
		}
		RemoveFromAskedEvent(SLA sla) { this.sla = sla; }
	}

	protected void bid_received(SLA bid) {}

	public void receive_ask(SLATP_Ask msg) {
		super.receive_ask(msg);

		int added_delay=ADDED_DELAY;
		SLABuyer buyer = (SLABuyer) msg.get_from();
		{
			if(buyer.get_node()!=node) {
				int iface = node.get_iface(buyer.get_node());
				LinkManager lm = node.get_iface(iface);
				if(lm==null) return;
				Link l = lm.get_link();

				added_delay+=l.get_max_delay();
			}
		}

		int  delay = msg.max_delay-added_delay;
		if(delay<MIN_DELAY) return;

		long exp   = msg.expiration+added_delay+GIVENSLA_EXPIRATION_MARGIN;

		// if it's for a neighbour destination, provision it
		// immediately.
		if(node.get_iface(msg.dest)>=0) {
			if(DEBUG_LAZY) s.debug(this,"ask received for neighbour.");
			buy_best(msg.dest,msg.bw,delay,exp);
		}
		else {
			ask_if_needed(buyer,msg.dest,msg.bw,delay,exp);
		}
	}

	int min_bw_increase(Node dest) {
		int total_bw=0;
		{
			Iterator it = search_taken_sla(dest,Integer.MAX_VALUE);
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				total_bw += sla.bw;
			}
		}

		int bw = (int)(total_bw*TAKENSLAS_MAX_INCREASE_BW);
		if(bw<TAKENSLAS_MIN_INCREASE) bw=TAKENSLAS_MIN_INCREASE;
		if(bw>TAKENSLAS_MAX_INCREASE) bw=TAKENSLAS_MAX_INCREASE;
		return bw;
	}

	protected void ask_if_needed(SLABuyer buyer, Node dest, int bw,
		int max_delay, long exp)
	{
		Iterator it;

		// verify if the residual bandwidth is enough to satisfy this
		// request
		it = search_taken_sla(dest,max_delay);
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			if(sla.expiration<exp) continue;
			if(sla.rbw>bw) return;
		}

		// didn't we already send such a request?
		it = asked.search_dest(dest).iterator();
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			if(sla.expiration<exp) continue;
			if(sla.max_delay>max_delay) continue;
			if(sla.bw<bw) continue;
			return;
		}

		// adjust expiration (we buy in big quantities only...)
		long now   = s.now();
		if(exp<now+ASK_MIN_EXPIRATION) exp=now+ASK_MIN_EXPIRATION;

		// adjust bandwidth
		int min_bw = min_bw_increase(dest);
		if(bw<min_bw) bw = min_bw;

		// put in the list of asked slas (generate a fake SLA)
		SLA sla = new SLA(dest,-1,bw,max_delay,-1,-1,exp,-2);
		sla.ds = ds_numbers.get_ds(-2,dest);
		asked.put(sla);
		// remove from list after ASK_WAIT_TIME
		s.schedule(ASK_WAIT_TIME+1, new RemoveFromAskedEvent(sla));

		// it is needed... -> ask it to everybody except buyer
		
		for(int i=0; i<peers; i++) {
			Service service = peer_service[i];
			if(service==null) continue;
			if(service.get_node()==node) continue;
			if(service==buyer) continue;

			send_ask((SLASeller) service,dest,bw,max_delay,exp);
		}

		// wait some time and buy it.

		s.schedule(ASK_WAIT_TIME, new BuyEvent(dest,bw,max_delay,exp));
	}

	protected void buy_best(Node dest, int bw, int max_delay, long exp) {
		if(DEBUG_LAZY) s.debug(this,"buy_best: dest="+dest+
			", bw="+bw+", max_delay="+max_delay);

		SLA bid = best_bid(dest, bw, 2*bw, max_delay, budget, exp);
		if(bid!=null) accept_bid(bid);
	}

	/**
	 * find bid with best price/volume
	 */
	protected SLA best_bid(Node dest, int min_bw, int max_bw,
		int max_dl, long max_cost, long exp)
	{
		//s.debug(this,"best_bid("+dest+","+max_bw+","+max_dl+","+max_cost+","+exp+")");
		// dest,bw,rbw,rbwbw,delay,cost,exp
		Iterator it=sla_map.search_taken_bids(dest,
			min_bw, max_bw+1, // bw
			-1, -1, // rbw
			-1, -1, // rbwbw
			-1, max_dl, // delay
			-1, max_cost, // cost
			exp, -1);

		SLA best=null;
		float min_cost=Float.MAX_VALUE;
		while(it.hasNext()) {
			SLA bid = (SLA) it.next();
			float cost_per_volume = (float) bid.cost_volume();
			if(cost_per_volume<min_cost) {
				min_cost=cost_per_volume;
				best = bid;
			}
		}

		return best;
	}

	public String toString() {
		return "SLATRouterLazy("+get_node()+")";
	}

	public SLATRouterLazy(Simulator s, Node node, int port,
		int local_peers, long budget, long fixed_costs, Random random)
	{
		super(s,node,port,local_peers,budget,fixed_costs,random);

		asked = new SLAList(s);
	}
}
