package flowsim.slatr;

import flowsim.core.*;

import java.util.Random;

import com.sun.java.util.collections.Set;
import com.sun.java.util.collections.HashSet;
import com.sun.java.util.collections.Iterator;

import java.io.*;

/* note: for the moment there is no differentiation on delay class */
public class SLATRouterProfitable extends SLATRouterImpl
{
	final boolean DEBUG_PROFITABLE;

	static final int CONNECTIVITY_MIN_BW = 64000;  // each destination must have
	                                               // at least that total bandwidth
	                                               // (if possible)
	static final int CONNECTIVITY_MAX_BW = 128000; // buy no more than
	                                               // this to ensure connectivity
	static final long CONNECTIVITY_WAIT_BEFORE_BUYING = 200000;
	static final boolean RBW_CONNECTIVITY = true;

	// bids with an expiration less than this won't be even considered.
	static final long TAKENBID_MIN_EXPIRATION = BASIS_CYCLE_TIME;

	// if residual bandwidth is less than this, try to buy some more
	static final float TAKENSLAS_MAX_RBWBW = 0.3f;
	// buy no more as 50% of already available bw.
	static final int   TAKENSLAS_MAX_INCREASE     = 512000;
	static final float TAKENSLAS_MAX_INCREASE_REL = 0.5f;
	static final int   TAKENSLAS_MIN_INCREASE     = 64000;

	// what is expected to remain unused. After that (more unused than
	// that), it isn't no more profitable.
	static final float TAKENSLAS_EXPECTED_RBWBW = 0.3f;
	// it is expected that the buyers are willing to buy at a price 20% higher
	// warning: if PRICE_TOLERANCE is higher than own price margin, a
	// service-loop can be made!
	//static final float GIVENSLAS_EXPECTED_PRICE_TOLERANCE = 1.2f;

	Set known_dests;

	class NewDestEvent implements Event {
		Node dest;
		public void trigger() {
			known_dests.add(dest);
			buy_dest_if_needed(dest);
		}
		NewDestEvent(Node d) { dest=d; }
	}

	// when a destination was previously unknown, wait some time and buy
	// the best bid to it.
	public void bid_received(SLA bid) {
		if(!known_dests.contains(bid.dest)) {
			s.schedule(CONNECTIVITY_WAIT_BEFORE_BUYING,
				new NewDestEvent(bid.dest));
		}
	}

	public void reject_received(SLA sla) {
		buy_dest_if_needed(sla.dest);
	}

	protected void buy_slas() {
		Iterator it = known_dests.iterator();
		while(it.hasNext()) {
			Node n = (Node) it.next();
			buy_dest_if_needed(n);
		}
	}

	/**
	 * find bid with best price/volume
	 */
	protected SLA best_bid(Node dest, int max_bw,
		int max_dl, long max_cost)
	{
		// dest,bw,rbw,rbwbw,delay,cost,exp
		Iterator it=sla_map.search_taken_bids(dest,
			-1, max_bw+1, // bw
			-1, -1, // rbw
			-1, -1, // rbwbw
			-1, max_dl, // delay
			-1, max_cost, // cost
			s.now()+TAKENBID_MIN_EXPIRATION, -1);

		SLA best=null;
		float min_cost=Float.MAX_VALUE;
		final long now = s.now();
		while(it.hasNext()) {
			SLA bid = (SLA) it.next();
			float cost_per_volume = (float) bid.cost_volume(now);
			if(cost_per_volume<min_cost) {
				min_cost=cost_per_volume;
				best = bid;
			}
		}

		return best;
	}

	protected void buy_connectivity(Node dest) {
		if(DEBUG_PROFITABLE) s.debug(this,"buy_connectivity("+dest+")");
		SLA bid = best_bid(dest, CONNECTIVITY_MAX_BW,
				Integer.MAX_VALUE, // max_delay
				Long.MAX_VALUE // max_cost
			);

		if(bid!=null) accept_bid(bid);
	}

	/**
	 * Is this bid a profitable one? The algorithm is as follows: take
	 * the average price of the given_slas which resulted from buying the
	 * last taken_sla. If this bid was bought and the expected amount
	 * (TAKENSLAS_EXPECTED_RBWBW) was sold at that price, would we make
	 * profit?
	 */

	protected boolean is_profitable(SLA bid) {
		// search last-bought taken_sla
		SLA last_taken_sla=null;
		{
			Iterator it=search_taken_sla(bid.dest,Integer.MAX_VALUE);
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				if(last_taken_sla==null ||
					last_taken_sla.time>sla.time)
				{
					last_taken_sla = sla;
				}
			}
		}
		if(last_taken_sla==null) {
			// uhmm... shouldn't this be handled in
			// buy_connectivity?
			s.debug(this, "Warning: is_profitable called on bid "+
					"for new destination.");
			return true;
		}

		// calculate average price of services sold using this
		// taken_sla
		/*
		float avg_price_volume;
		{
			long total_volume=0, total_cost=0;

			Iterator it=sla_map.given_slas_to_dest(bid.dest);
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				if(sla_map.get_taken_from_given(sla)!=last_taken_sla) continue;
				total_volume += sla.volume();
				total_cost   += sla.cost;
			}
			if(total_cost==0) avg_price_volume = Float.MAX_VALUE;
			else              avg_price_volume = (float)total_cost/total_volume;
		}
		*/

		long expected_volume = (long)(bid.volume(s.now())*(1.0-TAKENSLAS_EXPECTED_RBWBW));
		//float expected_price = avg_price_volume*GIVENSLAS_EXPECTED_PRICE_TOLERANCE;
		long expected_income = (long) price(bid.dest,expected_volume,bid.max_delay,bid.bw);

		if(bid.cost<expected_income) {
			//s.debug(this,"expected_volume="+expected_volume+
			//	" avg_price_volume="+avg_price_volume);

			if(DEBUG_PROFITABLE) {
				s.debug(this,"expected_income="+
					expected_income+" for "+bid);
			}

			return true;
		}
		else {
			s.debug(this,"not profitable: expected_income="+ expected_income+" for "+bid);
			return false;
		}
	}

	protected void buy_usage(Node dest, int max_bw) {
		if(DEBUG_PROFITABLE) s.debug(this,"buy_usage("+dest+
			", "+max_bw+") budget="+budget);

		SLA bid = best_bid(dest, max_bw,
				Integer.MAX_VALUE, // max_delay
				Long.MAX_VALUE // max_cost
			);

		if(bid!=null && is_profitable(bid)) accept_bid(bid);
	}

	int usage_max_bw_increase(int current_bw) {
		int bw = (int)(current_bw*TAKENSLAS_MAX_INCREASE_REL);
		if(bw<TAKENSLAS_MIN_INCREASE) bw=TAKENSLAS_MIN_INCREASE;
		if(bw>TAKENSLAS_MAX_INCREASE) bw=TAKENSLAS_MAX_INCREASE;
		return bw;
	}

	protected void buy_dest_if_needed(Node dest) {
		// count total bandwidth and residual-bandwidth to that
		// destination
		int total_bw=0;
		int total_rbw=0;
		{
			Iterator it = search_taken_sla(dest,Integer.MAX_VALUE);
			while(it.hasNext()) {
				SLA sla = (SLA) it.next();
				total_bw  += sla.bw;
				total_rbw += sla.rbw;
			}
		}

		// connectivity
		if(( RBW_CONNECTIVITY && total_rbw<CONNECTIVITY_MIN_BW) ||
		   (!RBW_CONNECTIVITY && total_bw<CONNECTIVITY_MIN_BW)) {
			buy_connectivity(dest);
			//return;
		}

		// usage
		if(total_rbw<total_bw*TAKENSLAS_MAX_RBWBW) {
			buy_usage(dest,usage_max_bw_increase(total_bw));
			return;
		}
	}
		
	/********************************************************/

	public String toString() {
		return "SLATRouterProf("+get_node()+")";
	}

	public SLATRouterProfitable(Simulator s, Node node, int port,
		int local_peers, long budget, long fixed_costs, Random random)
	{
		super(s,node,port,local_peers,budget,fixed_costs,random);

		known_dests = new HashSet();

		if(node.get_name().equals("5")) {
			DEBUG_PROFITABLE=true;
		}
		else {
			DEBUG_PROFITABLE=false;
		}
	}
}
