package flowsim.slatr;

import flowsim.core.*;

import java.util.Random;

import com.sun.java.util.collections.Collection;
import com.sun.java.util.collections.Iterator;

import java.io.*;

public class SLATRouterTrendy extends SLATRouterImpl
{
	final boolean DEBUG_TRENDY;

	// bids which will expire in less won't be accepted
	static final long ACCEPT_MARGIN       = 2*BASIS_CYCLE_TIME;

	static final int ACCEPT_MAX_BW = 128000;

	// if a taken sla is used at 80%, buy new one
	static final float TAKENSLAS_MAX_RBWBW = (float) 0.2;
	static final long  TAKENSLAS_MAX_EXP   = 10*BASIS_CYCLE_TIME;
	static final float SLA_MAX_INCREASE = 0.5f; // buy no more as 50% of already available bw.
	static final int SLA_MIN_BW_INCREASE = 128000;

	// if already sent bid is less than 50% more expensive, don't send
	// another one
	static final double  EQUALBID_PRICE_MARGIN = 1.5;

	public static ServiceFactory trendy_factory(final Simulator s, final int local_peers,
		final long budget, final long fixed_costs, final Random r)
	{
		return new ServiceFactory() {
			public Service create(Node node, int port) {
				return new SLATRouterTrendy(s,node,port,
					local_peers,budget,fixed_costs,r);
			}
		};
	}

	public void bid_received(SLA bid) {
		Iterator it;

		if(bid.expiration < s.now()+ACCEPT_MARGIN) return;
		if(bid.iface<0) return;
		if(bid.bw>ACCEPT_MAX_BW) return;
		if(bid.cost>budget) return;

		// can we already go to that dest?
		it=sla_map.search_taken_slas(bid.dest, -1, -1, -1, -1, -1, -1, -1, -1, -1,-1,
				s.now()+MIN_BID_EXPIRATION, -1);
		if(it.hasNext()) {
			if(node.get_iface(bid.dest)>=0 ||
				get_rbwbw_taken_sla_same_class(bid.dest,
						bid.max_delay)>TAKENSLAS_MAX_RBWBW) {
				if(DEBUG_TRENDY) s.debug(this,"A good sla is alread there...");
				return;
			}
		}

		// did we already send an accept for that destination?
		it=sla_map.search_taken_bids(bid.dest, -1, -1, -1, -1, -1, -1, -1, -1, -1,-1,
				s.now()+MIN_BID_EXPIRATION, -1);
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();
			if(sla.accepted) {
				if(DEBUG_TRENDY) s.debug(this,"Accept for similar bid already sent...");
				return;
			}
		}

		if(DEBUG_TRENDY) s.debug(this,"New dest node: "+bid.dest);

		if(links_rbw[bid.iface]>bid.bw) {
			accept_bid(bid);
		}
		else {
			if(DEBUG_TRENDY) s.debug(this,"Can't accept bid: link_rbw too small");
		}
	}
	
	final double rbwbw(SLA sla) {
		return (double) sla.rbw/sla.bw;
	}

	void buy_new_sla(SLA sla) {
		long total_bw = get_bw_taken_sla_same_class(sla.dest, sla.max_delay);
		int max_bw = (int) (total_bw * SLA_MAX_INCREASE);
		if(max_bw<SLA_MIN_BW_INCREASE) max_bw=SLA_MIN_BW_INCREASE;

		int dlc = dlclass(sla.max_delay);
		Iterator it = sla_map.search_taken_bids(sla.dest,
				-1, max_bw+1,
				-1, -1,
				-1, -1,
				-1, dlclass_max(dlc),
				-1L, -1L,
				s.now()+MIN_BID_EXPIRATION, -1);

		// find the less expensive (pro volume)
		SLA bid_to_buy=null;
		float min_cost=Float.MAX_VALUE;
		while(it.hasNext()) {
			SLA bid = (SLA) it.next();
			float cost_per_volume = (float) bid.cost_volume();
			if(cost_per_volume<min_cost) {
				min_cost=cost_per_volume;
				bid_to_buy = bid;
			}
		}

		if(bid_to_buy!=null) {
			if(DEBUG_TRENDY) s.debug(this,"wanted: "+sla);
			if(DEBUG_TRENDY) s.debug(this,"chosen: "+bid_to_buy);
			accept_bid(bid_to_buy);
		}
	}

	protected void buy_slas() {
		long exp = s.now()+TAKENSLAS_MAX_EXP;
		Iterator it;

		it = sla_map.search_taken_slas(null,
			-1,-1, // bw
			-1,-1, // rbw
			-1,TAKENSLAS_MAX_RBWBW,
			-1,-1, // delay
			-1L,-1L, // cost
			-1L,-1L // exp
		);
		while(it.hasNext()) {
			SLA sla = (SLA) it.next();

			// don't buy from neighbours if dest is neighbour...
			if(node.get_iface(sla.dest)>=0) continue;
		
			if(get_rbwbw_taken_sla_same_class(sla.dest,
				sla.max_delay)>TAKENSLAS_MAX_RBWBW) continue;
			if(DEBUG_TRENDY) s.debug(this, "taken sla "+sla+" full.");
			buy_new_sla(sla);
		}
	}

	public String toString() {
		return "SLATRouterTrendy("+get_node()+")";
	}

	public SLATRouterTrendy(Simulator s, Node node, int port,
		int local_peers, long budget, long fixed_costs, Random random)
	{
		super(s,node,port,local_peers,budget,fixed_costs,random);
		if(node.get_name().equals("0")) {
			DEBUG_TRENDY=true;
		}
		else {
			DEBUG_TRENDY=false;
		}

	}
}
