package flowsim.slatr;

import flowsim.core.*;

import com.sun.java.util.collections.Iterator;

/**
 * User of services provided by a SLATRouter.
 * @see SLATRouter
 */
abstract public class SLATUser extends UserImpl implements SLABuyer
{
	static final boolean DEBUG=false;
	static final int WAIT_FOR_BID_TIME = 1;
	static final int EXPIRATION_MARGIN = 1;
	static final int ACCEPT_TIME = 1;
	protected SLASeller server;

	long gain;

	SLAList bids;
	SLAList slas;
	SLAList buying;

	final class RemoveSlaEvent implements Event {
		final SLA sla;
		public void trigger() {
			remove_sla(sla);
		}
		RemoveSlaEvent(final SLA sla) { this.sla = sla; }
	}

	final class RemoveBidEvent implements Event {
		final SLA bid;
		public void trigger() {
			remove_bid(bid);
		}
		RemoveBidEvent(final SLA bid) { this.bid = bid; }
	}

	final class BuyBestBidEvent implements Event {
		final SLA sla;
		public void trigger() {
			buy_best_bid(sla);
		}
		BuyBestBidEvent(final SLA sla) { this.sla = sla; }
	}

	public SLABuyerProfile get_profile() {
		return new SLABuyerProfile(false,true,false);
	}

	public void remove_sla(SLA sla) {
		slas.remove(sla,null);
	}

	public void remove_bid(SLA bid) {
		bids.remove(bid,null);
	}

	public void bid_received(SLA bid) {}

	final public void receive_bid(SLATP_Bid msg) {
		if(msg.sla.dest==node) {
			s.error(this,"seller "+msg.get_from()+" sent me a bid to myself!");
			return; // It's me!
		}

		msg.sla.peer=0;
		msg.sla.iface=-1;
		msg.sla.accepted=false;

		SLA dup;
		if((dup=bids.get(0,msg.sla.dest,msg.sla.ds))!=null) {
			s.error(this,"received bid with already used DS-Byte!");
			s.error(this,"dup="+dup+", received="+msg.sla);
		}

		bids.put(msg.sla);

		if(DEBUG) s.debug(this,"Received bid "+msg.sla+" from "+msg.get_from());

		// schedule expiration
		msg.sla.expiration_event = s.schedule_absolute(
			msg.sla.expiration,new RemoveBidEvent(msg.sla));

		bid_received(msg.sla);
	}

	public void confirm_received(SLA sla) {
		utility += sla.utility;
		gain += sla.gain;
	}
	
	final public void receive_confirm(SLATP_Confirm msg) {
		if(DEBUG) s.debug(this,"Confirm received for sla ["+
			msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");

		SLA bid = bids.get(0,msg.get_sla_dest(),msg.get_sla_ds());

		// Check if an accept message was sent or if it is junk...
		if(!bid.accepted) {
			s.error(this,"Junk SLAT_Confirm: "+
				"["+msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");
			return;
		}

		// detract spent money
		remove_from_budget(bid.cost);

		// move to slas
		bids.remove(bid);
		slas.put(bid);

		buying.remove(bid);

		// re-schedule expiration in 'slas' instead of 'bids'
		if(bid.expiration_event!=null) {
			bid.expiration_event.remove();
		}
		bid.expiration_event = s.schedule_absolute(
			bid.expiration,new RemoveSlaEvent(bid));

		confirm_received(bid);
	}

	public final void receive_reject(SLATP_Reject msg) {
		if(DEBUG) s.debug(this,"Reject received for sla ["+
			msg.get_sla_dest()+", "+msg.get_sla_ds()+"]");

		SLA bid = bids.get(0,msg.get_sla_dest(),msg.get_sla_ds());

		// Check if an accept message was sent or if it is junk...
		if(!bid.accepted) {
			s.error(this,"Junk SLAT_Reject message received!");
			return;
		}

		bids.remove(bid,null);
		
		// try again
		SLA ideal = buying.get(0,msg.get_sla_dest(),msg.get_sla_ds());
		if(ideal==null) {
			s.debug(this,"receive_reject: ideal=null.");
		}
		else {
			buy_best_bid(ideal);
		}
	}

	void accept_bid(SLA bid) {
		if(DEBUG) s.debug(this,"Accepting received bid");
		send_accept(bid.dest,bid.ds);
	}

	void send_accept(Node dest, int ds) {
		if(DEBUG) {
			s.debug(this,"Sending accept for bid "+ "["+dest+", "+ds+"]");
		}
		
		SLA bid = bids.get(0,dest,ds);
		if(bid==null) {
			s.error(this,"Attempt to send accept of unknown bid "+
			"["+dest+", "+ds+"]");
			return;
		}
		if(bid.accepted==true) {
			if(DEBUG) s.debug(this,"Accept already sent for bid "+bid);
			return;
		}
		bid.accepted=true;
		
		SLATP_Accept.send(s,this,server,dest,ds);
	}

	/** definition:
	 *  bid_utility(ideal,ideal)=1, bid_utility(ideal,nothing)=0
	 */
	abstract public float bid_utility(SLA ideal, SLA bid);

	/** willingness to pay for maximal utility */
	abstract public int utility_wtp(SLA sla);

	protected void send_ask(Node dest, int bw, int max_delay, long expiration)
	{
		if(DEBUG) {
			s.debug(this,"Sending ask ["+dest+
					", bw="+bw+
					", dl="+max_delay+
					", exp="+expiration+"]");
		}
		SLATP_Ask.send(s,this,server,dest,bw,max_delay,expiration);
	}

	protected void ask(SLA ideal) {
		send_ask(ideal.dest, ideal.bw, ideal.max_delay, ideal.expiration);
	}

	protected void buy(SLA ideal) {
		max_utility+=1.0;

		ideal.expiration += WAIT_FOR_BID_TIME+ACCEPT_TIME+EXPIRATION_MARGIN;

		// ask it
		ask(ideal);

		// wait and try to buy best bid
		s.schedule(WAIT_FOR_BID_TIME, new BuyBestBidEvent(ideal));
	}

	protected boolean buy_best_bid(SLA ideal) {
		if(DEBUG) s.debug(this,"Searching for best bid (ideal="+ideal+")");

		Iterator it = bids.search_dest(ideal.dest).iterator();
		long max_gain=Long.MIN_VALUE;
		SLA   best_bid=null;

		while(it.hasNext()) {
			SLA bid = (SLA) it.next();
			//if(bid.cost>=budget) continue;
			long gain = (long) (bid_utility(ideal,bid)*utility_wtp(ideal)-bid.cost);
			if(DEBUG) s.debug(this,bid+" -> "+gain);
			if(gain>max_gain) {
				max_gain=gain;
				best_bid=bid;
			}
		}

		if(max_gain>0 && best_bid!=null) {
			best_bid.utility = bid_utility(ideal,best_bid);
			best_bid.gain = max_gain;
			ideal.ds = best_bid.ds;
			buying.put(ideal);
			accept_bid(best_bid);
			return true;
		}
		else {
			if(DEBUG) {
				s.debug(this, "No good bid found "+
					"(ideal="+ideal+", max_gain="+max_gain+
					", best_bid="+best_bid+")");
				/*
				SLATRouter r = (SLATRouter) server;
				r.debug_table();
				*/
			}
			return false;
		}
	}

	public String toString() {
		return "SLATUser("+get_node()+","+get_port()+")";
	}

	public void print_stats() {
		System.out.print(this+": budget="+budget+", budget_incr="+
				budget_incr+", gain="+gain+", max_utility="+max_utility
				+", utility="+utility);
	}

	/**
	 * @param s simulator object.
	 * @param node node where the user resides.
	 * @param port port where the user resides.
	 * @param budget initial budget.
	 * @param budget_incr budget increase pro second.
	 * @param server_port port of (unique) SLASeller for this user.
	 */
	public SLATUser(Simulator s, Node node, int port, long budget, long budget_incr,
		int server_port)
	{
		super(s,node,port,budget,budget_incr);

		this.slas = new SLAList(s,""+get_node()+","+get_port()+",s");
		this.bids = new SLAList(s,""+get_node()+","+get_port()+",b");
		this.buying = new SLAList(s,""+get_node()+","+get_port()+",buy");

		this.server = (SLASeller) node.get_service(server_port);
		if(this.server==null) {
			s.error(this,"server = null");
		}
		/* register to server */
		this.server.add_buyer(this,get_profile());
	}
}
