package flowsim.slatr;

import flowsim.core.*;
import flowsim.util.*;

import java.util.Random;

public class SLATUserVoice extends SLATUser implements FlowMonitor
{
	static final int BANDWIDTH_MIN  = 4*1000;  // 4 kbps
	static final int BANDWIDTH_GOOD = 16*1000; // 16 kbps
	static final int DELAY_GOOD     = 250*1000; // 50 ms (should be)
	static final int DELAY_MAX      = 500*1000; // 500 ms

	protected final Topology t;
	protected final Random r;
	protected final Distribution arrival;
	protected final Distribution duration;
	protected final NodeChooser to_chooser;
	protected final int calls_per_second;
	protected final int flow_id;
	protected final float wtp_per_volume;

	class MakeCallEvent implements Event {
		public void trigger() {
			make_call();
			s.schedule((long) arrival.rand(),new MakeCallEvent());
		}
	}

	final class StopFlowEvent implements Event {
		final Flow flow;
		public void trigger() {
			flow.stop();
		}
		StopFlowEvent(final Flow flow) { this.flow = flow; }
	}

	/** willingness to pay for maximal utility */
	public int utility_wtp(SLA sla) {
		return (int)(wtp_per_volume*sla.volume());
	}

	public void flow_started(Flow f) {
		if(DEBUG) s.debug(this,"Flow "+f+" started");
	}

	public void flow_stopped(Flow f) {
		if(DEBUG) s.debug(this,"Flow "+f+" stopped");
	}

	public void flow_blocked(Flow f) {
		s.error(this,"Flow "+f+" was blocked!!");
		((SLATRouter)server).debug_table();
	}

	public void flow_changed_bw(Flow f, int was_bw) {
		s.error(this,"Flow "+f+" changed bandwidth!!");
	}

	public void flow_changed_desired_bw(Flow f, int was_bw) {}
	public void flow_changed_available_bw(Flow f, int was_bw) {}

	public void confirm_received(SLA sla) {
		super.confirm_received(sla);

		long exp = sla.expiration-EXPIRATION_MARGIN;
		if(exp<s.now()) return; // already expired...

		RoutedFlow f = new UnicastFlowImpl(s, get_node(), sla.dest, sla.bw, flow_id,
			sla.ds, server.get_port());
		f.set_from_service((Service) this);
		f.attach_monitor(this);
		
		if(DEBUG) s.debug(this,"Starting flow "+f+" (stop="+exp+")");
		f.start();

		s.schedule_absolute(exp, new StopFlowEvent(f));
	}
	
	protected void ask(SLA ideal) {
		final int delta_bw = BANDWIDTH_GOOD-BANDWIDTH_MIN;
		final int delta_dl = DELAY_MAX-DELAY_GOOD;
		send_ask(ideal.dest, BANDWIDTH_GOOD, DELAY_GOOD, ideal.expiration);
		// FIXME: Uncomment the following!!!!! (testing with only one bid)
		//send_ask(ideal.dest, BANDWIDTH_GOOD-delta_bw/2, DELAY_GOOD, ideal.expiration);
		//send_ask(ideal.dest, BANDWIDTH_GOOD, DELAY_GOOD+delta_dl/2, ideal.expiration);
		//send_ask(ideal.dest, BANDWIDTH_GOOD-delta_bw/2, DELAY_GOOD+delta_dl/2, ideal.expiration);
	}

	void make_call() {
		// destination
		Node to;
		do {
			to = to_chooser.get();
		} while(get_node() == to);

		// duration
		int d = (int) duration.rand();

		SLA ideal = new SLA(to,-1,BANDWIDTH_GOOD,DELAY_GOOD,
			-1,s.now(),s.now()+d);
	
		buy(ideal);
	}

	float bid_utility_bw(SLA bid) {
		if(bid.bw>=BANDWIDTH_GOOD) return 1;
		if(bid.bw<=BANDWIDTH_MIN) return 0;
		return (float)(bid.bw-BANDWIDTH_MIN)/(BANDWIDTH_GOOD-BANDWIDTH_MIN);
	}

	float bid_utility_dl(SLA bid) {
		if(bid.max_delay>=DELAY_MAX) return 0;
		if(bid.max_delay<=DELAY_GOOD) return 1;
		return (float)(bid.max_delay-DELAY_GOOD)/(DELAY_MAX-DELAY_GOOD);
	}
	/** definition:
	 *  bid_utility(ideal,ideal)=1, bid_utility(ideal,nothing)=0
	 */
	public float bid_utility(SLA ideal, SLA bid) {
		if(ideal.dest!=bid.dest) return 0;
		if(ideal.expiration>bid.expiration) return 0;

		//System.out.println("utility="+bid_utility_bw(bid)*bid_utility_dl(bid));
		return bid_utility_bw(bid)*bid_utility_dl(bid);
	}

	/**
	 * @param node node where the user resides
	 * @param port port where the user resides
	 * @param budget_incr budget increase pro second
	 * @param server_port port of (unique) SLASeller for this user
	 * @param calls_per_second number of calls per second which should be generated (randomly)
	 * @param mean_duration average duration of calls in microseconds
	 * @param r Random object used to make random call distribution
	 */
	public SLATUserVoice(Simulator s, Topology t, Node node, int port, long budget,
		long budget_incr, int server_port, int calls_per_second,
		int mean_duration, Random r, NodeChooser to_chooser, int flow_id,
		float wtp_per_volume)
	{
		super(s,node,port,budget,budget_incr,server_port);
		this.r = r;
		this.arrival = new ExponentialDistribution((double)calls_per_second/1000000,r);
		this.duration = new ExponentialDistribution(1.0/mean_duration,r);
		this.to_chooser = to_chooser;
		this.t = t;
		this.calls_per_second = calls_per_second;
		this.flow_id = flow_id;
		this.wtp_per_volume = wtp_per_volume;

		s.schedule((long) arrival.rand(),new MakeCallEvent());

		if(DEBUG) s.debug(this,"SLATUserVoice instantiated");
	}
}
