package flowsim.slatr;

import flowsim.core.*;
import flowsim.dv.*;

import java.util.Random;

final class RandomSLATRouterFactory implements ServiceFactory
{
	static final float TRENDY = (float) 0; // 0..0.5 -> trendy
	static final float PROFI = (float) 1;   // 0.5..1 -> profitable

	final Simulator s;
	final int local_peers;
	final long budget;
	final long fixed_costs;
	final Random r;

	public Service create(Node node, int port) {
		if(r.nextFloat()<TRENDY) {
			return new SLATRouterTrendy(s,node,port,local_peers,budget,fixed_costs,r);
		}
		else {
			return new SLATRouterProfitable(s,node,port,local_peers,budget,fixed_costs,r);
		}
	}

	RandomSLATRouterFactory(final Simulator s, final int local_peers,
		final long budget, final long fixed_costs, final Random r)
	{
		this.s=s;
		this.local_peers=local_peers;
		this.budget=budget;
		this.fixed_costs=fixed_costs;
		this.r = r;
	}
}

final public class TestSLATR
{
	static Simulator s;
	static Topology t;
	static LinkFactory lf;
	static LinkManagerFactory lmf;
	static Random r;
	static final int NODES = 12;
	static final int MAX_NEIGHBOURS = 4;
	static final int DVROUTER_PORT = 0;
	static final int SLATROUTER_PORT = 1;
	static final int MAX_ROUTERS = 2;
	static final int USER_PORT = MAX_ROUTERS;
	static final int MAX_USERS=5;
	static int[] user_port;

	static void init() {
		r = new Random(12444);
		s = new Simulator();
		s.set_verbose(true);
		t = new Topology(s,NODES,MAX_NEIGHBOURS,MAX_ROUTERS+MAX_USERS,"logs");
		//t = new Topology(s,NODES,MAX_NEIGHBOURS,MAX_ROUTERS+MAX_USERS,null);

		t.add_services(DVRouterImpl.factory(s),DVROUTER_PORT);
		ServiceFactory sf = new RandomSLATRouterFactory(s, MAX_USERS, 1000000, 100000, r);
		t.add_services(sf,SLATROUTER_PORT);

		user_port = new int[NODES];
		for(int n=0; n<NODES; n++) {
			user_port[n]=USER_PORT;
		}
	}

	static void add_user_voice(int n, long budget,
		long budget_incr, int calls_per_second,
		int mean_duration)
	{
		int port = user_port[n]++;
		new SLATUserVoice(s, t, t.node(n), port, budget, budget_incr,
			SLATROUTER_PORT, calls_per_second, mean_duration, r,
			t.get_random_node_chooser(r),0,10);
	}

	static void add_user_ftp(int n, long budget,
		long budget_incr, int flow_id, int transfers_per_second,
		int average_size)
	{
		int port = user_port[n]++;
		NodeChooser nc = t.get_random_node_chooser(r);
		new FTPUser(s, t.node(n), port, nc, budget,
			budget_incr, flow_id, DVROUTER_PORT,
			transfers_per_second, average_size, r);
	}

	static void add_users() {
		add_user_voice(0,100000,100000,1000,500000);
		add_user_voice(0,100000,200000,1000,500000);
		add_user_voice(0,100000,300000,1000,500000);
		/*
		add_user_ftp(0,10000,1000,0,100,150000);
		add_user_ftp(1,10000,1000,1,100,150000);
		add_user_ftp(2,10000,1000,2,100,150000);
		*/
	}

	public static void main(String[] args) {
		init();

		// create links
		lf =  FixedDelayLink.randomdelay_factory(s,10000000,10000,20000,0);
		lmf = FairLinkManager.factory(s);
		t.add_duplex_link(0,1,lf,lmf);
		t.add_duplex_link(0,4,lf,lmf);
		t.add_duplex_link(0,9,lf,lmf);
		t.add_duplex_link(1,2,lf,lmf);
		t.add_duplex_link(1,6,lf,lmf);
		t.add_duplex_link(2,3,lf,lmf);
		t.add_duplex_link(2,4,lf,lmf);
		t.add_duplex_link(3,7,lf,lmf);
		t.add_duplex_link(4,5,lf,lmf);
		t.add_duplex_link(4,8,lf,lmf);
		t.add_duplex_link(5,6,lf,lmf);
		t.add_duplex_link(6,7,lf,lmf);
		t.add_duplex_link(7,9,lf,lmf);
		t.add_duplex_link(7,11,lf,lmf);
		t.add_duplex_link(8,9,lf,lmf);
		t.add_duplex_link(9,10,lf,lmf);
		t.add_duplex_link(10,11,lf,lmf);

		// create nam log
		//t.make_nam_log("testslatr.nam");

		// run simulation for some time
		s.run(1*1000*1000); // 1 seconds

		//print_routing_tables_slatr();

		add_users();

		// run simulation for some time
		s.run(5*1000*1000);

		// debug tables
		print_routing_tables_dv();
		print_routing_tables_slatr();

		print_user_stats();

		t.flush_logs();
	}

	static void print_routing_tables_slatr() {
		for(int i=0; i<t.nodes(); i++) {
			System.out.println();
			SLATRouter r = (SLATRouter) t.node(i).get_service(SLATROUTER_PORT);
			r.debug_table();
		}
	}

	static void print_routing_tables_dv() {
		for(int i=0; i<t.nodes(); i++) {
			System.out.println();
			DVRouter r = (DVRouter) t.node(i).get_service(DVROUTER_PORT);
			r.debug_table();
		}
	}

	static void print_user_stats() {
		// print routing tables
		for(int i=0; i<NODES; i++) {
			for(int u=0; u<MAX_USERS; u++) {
				User user = (User) t.node(i).get_service(USER_PORT+u);
				if(user!=null) {
					System.out.println();
					user.print_stats();
				}
			}
		}
	}
}
