
package flowsim.slatr;

import flowsim.core.*;
import flowsim.dv.*;

import java.util.Random;
import java.io.*;

final public class TestUserComp
{
	public static void main(String[] args) {
		TestUserCompRun test = new TestUserCompRun();
		test.run();
	}
}

final class TestUserCompRun
{
	Simulator s;
	Topology t;
	LinkFactory lf;
	LinkManagerFactory lmf;
	Random r;

	static final int     RUN_SECONDS = 50;
	static final boolean TEST_SLAT   = true; // if false: test DV

	static final int MAX_NEIGHBOURS = 2;
	static final int ROUTER_PORT = 0;
	static final int MAX_ROUTERS = 1;
	static final int USER_PORT = MAX_ROUTERS;
	static final int MAX_USERS=3;

	static final int CALLS_PER_SECOND   = 500;
	static final int CALL_BW            = 16000;
	static final int USER_BUDGET        = 100000;
	static final int ROUTER_BUDGET      = 1000000;
	static final int ROUTER_BUDGET_INCR = 1000000;

	static final boolean PERIODIC_DEBUG     = true;
	static final long PERIODIC_DEBUG_PERIOD = 1000000;
	static final long LOG_UTILITY_PERIOD    = 200000;

	static final int NODES = 3;

	float last_max_utility1=0, last_utility1=0;
	float last_max_utility2=0, last_utility2=0;
	float last_max_utility3=0, last_utility3=0;
	PrintStream logutility, logmaxutility;

	File logs_dir;
	int[] user_port;

	static final float wtp_volume1 = 0.05f;
	static final float wtp_volume2 = 0.08f;
	static final float wtp_volume3 = 0.5f;
	static final int budget_incr1 = 1000;
	static final int budget_incr2 = 1000;
	static final int budget_incr3 = 1000;
	static final int BANDWIDTH = 5000000;

	final class SLATRouterFactory implements ServiceFactory
	{
		static final float TRENDY = (float) 0;
		static final float PROFI  = (float) 1;
		static final float LAZY   = (float) 1;

		public Service create(Node node, int port) {
			float sel = r.nextFloat();

			if(sel<TRENDY) {
				return new
					SLATRouterTrendy(s,node,port,MAX_USERS,
							ROUTER_BUDGET,0,r);
			}
			else if(sel<PROFI) {
				return new
					SLATRouterProfitable(s,node,port,MAX_USERS,
							ROUTER_BUDGET,0,r);
			}
			else {
				return new
					SLATRouterLazy(s,node,port,MAX_USERS,
							ROUTER_BUDGET,0,r);
			}
		}
	}

	final class LogUtility implements Event {
		public void trigger() { log_utility(); }
	}

	void log_utility() {
		User u1, u2, u3;

		if(t.node(0)!=null) {
			u1 = (User) t.node(0).get_service(USER_PORT);
			u2 = (User) t.node(0).get_service(USER_PORT+1);
			u3 = (User) t.node(0).get_service(USER_PORT+2);
		}
		else {
			u1=u2=u3=null;
		}

		if(u1!=null && u2!=null && u3!=null) {
			float utility1     = u1.get_utility();
			float utility2     = u2.get_utility();
			float utility3     = u3.get_utility();

			float max_utility1 = u1.get_max_utility();
			float max_utility2 = u2.get_max_utility();
			float max_utility3 = u3.get_max_utility();

			long now = s.now();
			logutility.println(now
				+"\t"+(utility1-last_utility1)
				+"\t"+(utility2-last_utility2)
				+"\t"+(utility3-last_utility3)
			);
			logmaxutility.println(now
				+"\t"+(max_utility1-last_max_utility1)
				+"\t"+(max_utility2-last_max_utility2)
				+"\t"+(max_utility3-last_max_utility3)
			);

			last_utility1 = utility1; last_max_utility1 = max_utility1;
			last_utility2 = utility2; last_max_utility2 = max_utility2;
			last_utility3 = utility3; last_max_utility3 = max_utility3;
		}

		s.schedule(LOG_UTILITY_PERIOD,new LogUtility());
	}

	void init() {
		r = new Random(12444);
		s = new Simulator();
		s.set_verbose(true);
		logs_dir = new File("logs-uc");
		logs_dir.mkdir();

		user_port = new int[NODES];
		for(int n=0; n<NODES; n++) {
			user_port[n]=USER_PORT;
		}
		
		// create grid topology (nodes, links and linkmanagers)
		lf  =  FixedDelayLink.factory(s,10000000,10000);
		lmf = FairLinkManager.factory(s);

		t   = new Topology(s,NODES,MAX_NEIGHBOURS,MAX_ROUTERS+MAX_USERS,
				logs_dir.toString());

		ServiceFactory sf = new SLATRouterFactory();
		t.add_services(sf,ROUTER_PORT);

		t.add_duplex_link(0,1,lf,lmf);
		t.add_duplex_link(1,2,lf,lmf);

		// logs
		try {
			logutility = new PrintStream(
					new FileOutputStream(logs_dir.toString()+"/utility"));
			logmaxutility = new PrintStream(
					new FileOutputStream(logs_dir.toString()+"/max_utility"));
			
			log_utility();
		}
		catch(IOException e) {
			logutility=logmaxutility=null;
		}
	}

	void add_user_voice(int n, long budget,
		long budget_incr, int calls_per_second,
		int mean_duration,NodeChooser nodechooser,
		int flowid, float wtp_volume)
	{
		int port = user_port[n]++;
		if(TEST_SLAT) {
			new SLATUserVoice(s, t, t.node(n), port, budget, budget_incr,
					ROUTER_PORT, calls_per_second, mean_duration, r,
					nodechooser,flowid,wtp_volume);
		}
		else {
			new UserVoice(s, t, t.node(n), port, 
					ROUTER_PORT, calls_per_second, mean_duration, r,
					nodechooser,flowid);
		}
	}

	void add_users() {
		long size = BANDWIDTH/CALLS_PER_SECOND;
		int duration = (int)(size*1000000/CALL_BW);
		//int budget_incr = 10*CALLS_PER_SECOND+bw;
		s.debug(this,"size="+size+", duration="+duration);

		NodeChooser nodechooser1 = new NodeChooser() {
			public Node get() {
				return t.node(2);
			}
		};
		
		add_user_voice(0,USER_BUDGET,budget_incr1,
			CALLS_PER_SECOND,duration,nodechooser1,3,wtp_volume1);
		add_user_voice(0,USER_BUDGET,budget_incr2,
			CALLS_PER_SECOND,duration,nodechooser1,4,wtp_volume2);
		add_user_voice(0,USER_BUDGET,budget_incr3,
			CALLS_PER_SECOND,duration,nodechooser1,5,wtp_volume3);
	}

	void print_routing_tables_slatr(PrintStream ps) {
		for(int i=0; i<t.nodes(); i++) {
			ps.println();
			SLATRouter r = (SLATRouter) t.node(i).get_service(ROUTER_PORT);
			r.debug_table(ps);
		}
	}

	class PeriodicDebugEvent implements Event {
		public void trigger() { periodic_debug(); }
	}

	void periodic_debug() {
		try {
			FileOutputStream fos = new FileOutputStream("pdebug/pdebug-"+s.now());
			PrintStream ps = new PrintStream(fos);
			print_routing_tables_slatr(ps);
			s.schedule(PERIODIC_DEBUG_PERIOD,new PeriodicDebugEvent());
		}
		catch(IOException e) {}
	}

	public void run() {
		init();

		// create nam log
		//t.make_nam_log(logs_dir.toString()+"/testlb3.nam",NAM_CYCLE_TIME,true);

		if(PERIODIC_DEBUG && TEST_SLAT) periodic_debug();

		s.run(2*1000*1000); // 2 seconds

		add_users();

		s.run(RUN_SECONDS*1000*1000);

		//print_routing_tables_dv();
		print_routing_tables_slatr(System.out);

		t.flush_logs();
	}
}
